import errno
import fcntl
import numbers
import os
import pickle
import tempfile
import warnings
from struct import calcsize, pack, unpack

from pfio import cache
from pfio.cache.file_cache import _check_local, _default_cache_path


class _NoOpenNamedTemporaryFile(object):
    """Temporary file class

    This class warps mkstemp and implements auto-clean mechanism.
    The reason why we cannot use the tempfile.NamedTemporaryFile is that
    it has an unpicklable member because it opens the created temporary file,
    which makes it impossible to pass over to worker processes.

    The auto cleanup mechanism is based on CPython tempfile implementation.
    https://github.com/python/cpython/blob/3.8/Lib/tempfile.py#L406-L446
    """

    # Set here since __del__ checks it
    name = None
    master_pid = None

    def __init__(self, dir, master_pid):
        _, self.name = tempfile.mkstemp(dir=dir)
        self.master_pid = master_pid

    def close(self, unlink=os.unlink, getpid=os.getpid):
        if self.name and self.master_pid == getpid():
            unlink(self.name)
            self.name = None

    def __del__(self):
        self.close()


class _DummyTemporaryFile(object):
    """Dummy tempfile class that imitates the _NoOpenNamedTemporaryFile

    This class is used for MultiprocessFileCache.preload.
    The cache file fed from outside shouldn't be automatically deleted
    by close(), so it uses this dummy cache class.
    """

    def __init__(self, name):
        self.name = name

    def close(self):
        pass


class MultiprocessFileCache(cache.Cache):
    '''The Multiprocess-safe cache system on a local filesystem

    Stores cache data in a local temporary file, created in ``~/.pfio/cache``
    by default. It automatically deletes the cache data after the object is
    collected. When this object is not correctly closed (e.g., the process
    killed by SIGKILL), the cache remains after the process's death.

    This class supports handling a cache from multiple processes.
    A MultiprocessFileCache object can be handed over to another process
    through the pickle. Calling ``get`` and ``put`` in each process will
    look into the same cache file with flock-based locking. The temporary
    cache file will persist as long as the MultiprocessFileCache object is
    alive in the original process that creates it.
    Therefore, even after destroying the worker processes,
    the MultiprocessFileCache object can still be passed to another process.

    .. admonition:: Example

       Using MultiprocessFileCache is similar to the :class:`~NaiveCache`
       and :class:`~FileCache`.
       ::

           from pfio.cache import MultiprocessFileCache

           class MyDataset(torch.utils.data.Dataset):
               def __init__(self, image_paths):
                   self.paths = image_paths
                   self.cache = MultiprocessFileCache(len(image_paths), do_pickle=True)

               ...

       When iterating over the dataset, it is common to load the data
       concurrently to hide file IO bottleneck by setting higher ``num_workers``
       in PyTorch DataLoader.
       https://pytorch.org/docs/stable/data.html
       ::

           image_paths = open('/path/to/image_list.txt').read().splitlines()
           dataset = MyDataset(image_paths)
           loader = DataLoader(dataset, batch_size=64, num_workers=8)  # Parallel data loading

           for epoch in range(10):
               for batch in loader:
                   ...

       In this case, the dataset is distributed to each worker process
       i.e., ``__getitem__`` of the dataset will be called in a different
       process that initialized it.
       The ``MultiprocessFileCache`` object held by the dataset in each worker
       looks at the same cache file and handles the concurrent access based on
       the ``flock`` system call.
       Therefore the data inserted to the cache by a worker process
       can be accessed from another worker process safely.

       In case your task does not require concurrent data loading,
       i.e., ``num_workers=0`` in DataLoader, consider using :class:`~FileCache`
       as it has less overhead for concurrency control.

       The persisted cache file created by ``preserve()`` can be used for
       :meth:`FileCache.preload` and vice versa.

    Arguments:
        length (int): Length of the cache array.

        do_pickle (bool):
            Do automatic pickle and unpickle inside the cache.

        dir (str): The path to the directory to place cache data in
            case home directory is not backed by fast storage device.
            Must not be an NFS.

        cache_size_limit (None or int): Limitation of the cache size in bytes.
            If the total amount of cached data reaches the limit,
            the cache will become frozen and no longer acccept further addition.
            Data already stored in the cache can be accessed normally.
            None (default) and 0 is unlimited.

        verbose (bool):
            Print detailed logs of the cache.

    '''  # NOQA

    def __init__(self, length, do_pickle=False,
                 dir=None, cache_size_limit=None, verbose=False):
        self.length = length
        self.do_pickle = do_pickle
        self.verbose = verbose
        if self.length <= 0 or (2 ** 64) <= self.length:
            raise ValueError("length has to be between 0 and 2^64")

        if not (cache_size_limit is None or
                (isinstance(cache_size_limit, numbers.Number) and
                 0 <= cache_size_limit)):
            msg = "cache_size_limit has to be either None, zero " \
                  "(both indicate unlimited) or larger than 0. " \
                  "{} is specified.".format(cache_size_limit)
            raise ValueError(msg)
        self.cache_size_limit = cache_size_limit

        if dir is None:
            self.dir = _default_cache_path()
        else:
            self.dir = dir
        os.makedirs(self.dir, exist_ok=True)
        _check_local(self.dir)

        self.closed = False
        self._frozen = False
        self._master_pid = os.getpid()
        self.cache_file = _NoOpenNamedTemporaryFile(self.dir, self._master_pid)
        cache_fd = os.open(self.cache_file.name, os.O_RDWR)

        if self.verbose:
            print('created cache file:', self.cache_file.name)

        try:
            fcntl.flock(cache_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)

            # Fill up indices part of the cache file by index=0, size=-1
            buf = pack('Qq', 0, -1)
            self.buflen = calcsize('Qq')
            assert self.buflen == 16
            for i in range(self.length):
                offset = self.buflen * i
                r = os.pwrite(cache_fd, buf, offset)
                assert r == self.buflen
        except OSError as ose:
            # Lock acquisition error -> No problem, since other worker
            # should be already working on it
            if ose.errno not in (errno.EACCES, errno.EAGAIN):
                raise
        finally:
            fcntl.flock(cache_fd, fcntl.LOCK_UN)
            os.close(cache_fd)

        # Open lazily at the first call of get or put in each child process
        self._fd_pid = None
        self.cache_fd = None

    def __len__(self):
        return self.length

    @property
    def multiprocess_safe(self) -> bool:
        return True

    @property
    def multithread_safe(self) -> bool:
        return True

    def get(self, i):
        if self.closed:
            return
        data = self._get(i)
        if self.do_pickle and data:
            data = pickle.loads(data)
        return data

    def _open_fds(self):
        pid = os.getpid()
        if self._fd_pid != pid:
            self._fd_pid = pid
            self.cache_fd = os.open(self.cache_file.name, os.O_RDWR)

    def _get(self, i):
        if i < 0 or self.length <= i:
            raise IndexError("index {} out of range ([0, {}])"
                             .format(i, self.length - 1))

        self._open_fds()
        offset = self.buflen * i
        fcntl.flock(self.cache_fd, fcntl.LOCK_SH)
        index_entry = os.pread(self.cache_fd, self.buflen, offset)
        (o, l) = unpack('Qq', index_entry)
        if l < 0 or o < 0:
            fcntl.flock(self.cache_fd, fcntl.LOCK_UN)
            return None

        data = os.pread(self.cache_fd, l, o)
        assert len(data) == l
        fcntl.flock(self.cache_fd, fcntl.LOCK_UN)
        return data

    def put(self, i, data):
        if self._frozen or self.closed:
            return False

        try:
            if self.do_pickle:
                data = pickle.dumps(data)
            return self._put(i, data)

        except OSError as ose:
            # Disk full (ENOSPC) possibly by cache; just warn and keep running
            if ose.errno == errno.ENOSPC:
                warnings.warn(ose.strerror, RuntimeWarning)
                return False
            else:
                raise ose

    def _put(self, i, data):
        if self.closed:
            return False
        if i < 0 or self.length <= i:
            raise IndexError("index {} out of range ([0, {}])"
                             .format(i, self.length - 1))

        self._open_fds()

        index_ofst = self.buflen * i
        fcntl.flock(self.cache_fd, fcntl.LOCK_EX)
        buf = os.pread(self.cache_fd, self.buflen, index_ofst)
        (o, l) = unpack('Qq', buf)

        if l >= 0 and o >= 0:
            # Already data exists
            fcntl.flock(self.cache_fd, fcntl.LOCK_UN)
            return False

        data_pos = os.lseek(self.cache_fd, 0, os.SEEK_END)
        if self.cache_size_limit:
            if self.cache_size_limit < (data_pos + len(data)):
                self._frozen = True
                fcntl.flock(self.cache_fd, fcntl.LOCK_UN)
                return False

        index_entry = pack('Qq', data_pos, len(data))
        assert os.pwrite(self.cache_fd, index_entry, index_ofst) == self.buflen
        assert os.pwrite(self.cache_fd, data, data_pos) == len(data)
        os.fsync(self.cache_fd)
        fcntl.flock(self.cache_fd, fcntl.LOCK_UN)
        return True

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self):
        pid = os.getpid()

        if pid == self._fd_pid:
            os.close(self.cache_fd)
            self._fd_pid = None

        if not self.closed and pid == self._master_pid:
            self.cache_file.close()
            self.closed = True
            self.cache_file = None
            self.cache_fd = None

    def preload(self, name):
        '''Load the cache saved by ``preserve()``

        After loading the file, no data can be added to the cache.
        ``name`` is the name of the persistent file in the cache directory.

        When it succeeds, it returns ``True``.
        If there is no cache file with the specified name in
        the cache directory, it will do nothing but return ``False``.

        Be noted that ``preload()`` can be called only by the master process
        i.e., the process where ``__init__()`` is called,
        in order to prevent inconsistency.
        When using in a multiprocessing environment, you first need to create
        a ``MultiprocessFileCache`` object, call its ``preload()`` and then
        pass it to the worker processes.

        Returns:
            bool: Returns True if succeed.

        .. note:: This feature is experimental.

        '''
        if self._frozen:
            if self.verbose:
                print("Failed to preload the cache from {}: "
                      "The cache is already frozen."
                      .format(name))
            return False

        if self._master_pid != os.getpid():
            raise RuntimeError("Cannot preload a cache in a worker process")

        # Overwrite the current cache by the specified cache file.
        # This is needed to prevent the specified cache file are deleted when
        # the cache object is destroyed.
        ld_cache_file = os.path.join(self.dir, name)
        if not os.path.exists(ld_cache_file):
            if self.verbose:
                print('Failed to ploread the cache from {}: '
                      'The specified cache not found in {}'
                      .format(name, self.dir))
            return False

        self.cache_file.close()
        self.cache_fd = None
        self.cache_file = _DummyTemporaryFile(ld_cache_file)
        self._frozen = True
        return True

    def preserve(self, name, overwrite=False):
        '''Preserve the cache as a persistent file on the disk

        Once the cache is preserved, the cache file will not be removed
        at cache close. To read data from preserved file, use
        ``preload()`` method. After preservation, no data can be added
        to the cache. ``name`` is the name of the persistent files saved into
        the cache directory.

        When it succeeds, it returns ``True``.
        If there is a cache file with the same name already exists in the
        cache directory, it will do nothing but return ``False``.

        Be noted that ``preserve()`` can be called only by the master process
        i.e., the process where ``__init__()`` is called,
        in order to prevent inconsistency.

        The preserved cache can also be preloaded by :class:`~FileCache`.

        Arguments:
            name (str): Prefix of the preserved file names.
                ``(name).cachei`` and ``(name).cached`` are created.
                The files are created in the same directory as the cache
                (``dir`` option to ``__init__``).

            overwrite (bool): Overwrite if already exists

        Returns:
            bool: Returns True if succeed.

        .. note:: This feature is experimental.

        '''

        if self._master_pid != os.getpid():
            raise RuntimeError("Cannot preserve a cache in a worker process")

        cache_file = os.path.join(self.dir, name)
        if overwrite:
            if os.path.exists(cache_file):
                os.unlink(cache_file)
        elif os.path.exists(cache_file):
            if self.verbose:
                print('Specified cache named "{}" already exists in {}'
                      .format(name, self.dir))
            return False

        self._open_fds()
        try:
            fcntl.flock(self.cache_fd, fcntl.LOCK_EX)
            os.link(self.cache_file.name, cache_file)

        except OSError as ose:
            # Lock acquisition error -> No problem, since other worker
            # should be already working on it
            if ose.errno not in (errno.EACCES, errno.EAGAIN):
                raise
        finally:
            fcntl.flock(self.cache_fd, fcntl.LOCK_UN)

        return True



import os
import random
import string
import subprocess
from unittest import mock
from zipfile import ZipFile


class ZipForTest:
    def __init__(self, destfile, data=None):
        if data is None:
            self.data = dict(
                file=b"foo",
                dir=dict(
                    f=b"bar"
                )
            )
        else:
            self.data = data

        self._make_zip(destfile)
        self.destfile = destfile

    def content(self, path):
        d = self.data

        for node in path.split(os.path.sep):
            d = d.get(node)
            if not isinstance(d, dict):
                return d

    def _make_zip(self, destfile):
        with ZipFile(destfile, "w") as z:
            stack = []
            self._write_zip_contents(z, stack, self.data)

    def _write_zip_contents(self, z, stack, data):
        for k in data:
            if isinstance(data[k], dict):
                self._write_zip_contents(z, stack+[k], data[k])
            else:
                path = os.path.join(*stack, k)
                with z.open(path, 'w') as fp:
                    fp.write(data[k])


def make_zip(zipfilename, root_dir, base_dir):
    pwd = os.getcwd()
    with ZipFile(zipfilename, "w") as f:
        try:
            os.chdir(root_dir)
            for root, dirs, filenames in os.walk(base_dir):
                for _dir in dirs:
                    path = os.path.normpath(os.path.join(root, _dir))
                    f.write(path)
                for _file in filenames:
                    path = os.path.normpath(os.path.join(root, _file))
                    f.write(path)
        finally:
            os.chdir(pwd)


def make_random_str(n):
    return ''.join([random.choice(string.ascii_letters + string.digits)
                    for i in range(n)])


def randstring(length=16):
    letters = string.ascii_letters + string.digits
    return (''.join(random.choice(letters) for _ in range(length)))


def patch_subprocess(stdout, stderr=b''):
    def decorator(f):
        def wrapper(*args, **kwargs):
            orig_method = subprocess.run
            try:
                cp = subprocess.CompletedProcess(args='hoge', returncode=0)
                cp.stdout = stdout
                cp.stderr = stderr
                subprocess.run = mock.create_autospec(subprocess.run,
                                                      return_value=cp)
                return f(*args, **kwargs)
            finally:
                subprocess.run = orig_method

        return wrapper
    return decorator



import hashlib
import io
import os
from types import TracebackType
from typing import Optional, Type

import boto3
from botocore.exceptions import ClientError

from .fs import FS, FileStat

DEFAULT_MAX_BUFFER_SIZE = 16 * 1024 * 1024


def _normalize_key(key: str) -> str:
    key = os.path.normpath(key)
    if key.startswith("/"):
        return key[1:]
    else:
        return key


class S3ObjectStat(FileStat):
    def __init__(self, key, head):
        self.filename = key
        self.last_modified = head['LastModified'].timestamp()
        self.size = head['ContentLength']
        self.metadata = head['Metadata']
        self._head = head

    def isdir(self):
        return False


class S3PrefixStat(FileStat):
    def __init__(self, key):
        self.filename = key
        self.last_modified = 0
        self.size = -1

    def isdir(self):
        return True


class _ObjectReader(io.RawIOBase):
    def __init__(self, client, bucket, key, mode, kwargs):
        super(_ObjectReader, self).__init__()

        self.client = client
        self.bucket = bucket
        self.key = key

        res = self.client.head_object(Bucket=bucket, Key=key)
        if res.get('DeleteMarker'):
            raise FileNotFoundError()

        self._mode = mode
        self.pos = 0
        self.content_length = res['ContentLength']
        self._closed = False

    def read(self, size=-1) -> bytes:
        # Always returns binary; as this object is wrapped with
        # TextIOWrapper in case of text mode open.

        s = self.pos

        if self.pos >= self.content_length:
            return b''
        elif size <= 0:
            e = ''
        else:
            e = min(self.pos + size, self.content_length)

        r = 'bytes={}-{}'.format(s, e)
        res = self.client.get_object(Bucket=self.bucket,
                                     Key=self.key,
                                     Range=r)
        body = res['Body']

        if size < 0:
            data = body.read()
        else:
            data = body.read(size)

        self.pos += len(data)

        return data

    def readline(self):
        raise NotImplementedError()

    def close(self):
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        self.close()

    def flush(self):
        pass

    @property
    def closed(self):
        return self._closed

    def isatty(self):
        return False

    def readable(self):
        return True

    def seekable(self):
        return True

    def tell(self):
        return self.pos

    def truncate(self, size=None):
        raise io.UnsupportedOperation('truncate')

    def seek(self, pos, whence=io.SEEK_SET):
        if whence in [0, io.SEEK_SET]:
            if pos < 0:
                raise OSError(22, "[Errno 22] Invalid argument")
        elif whence in [1, io.SEEK_CUR]:
            pos += self.pos
        elif whence in [2, io.SEEK_END]:
            pos += self.content_length
        else:
            raise ValueError('Wrong whence value: {}'.format(whence))

        if pos < 0:
            raise OSError(22, "[Errno 22] Invalid argument")
        self.pos = pos
        return self.pos

    def writable(self):
        return False

    def write(self, data):
        raise io.UnsupportedOperation('not writable')

    def readall(self):
        return self.read(-1)

    def readinto(self, b):
        buf = self.read(len(b))
        b[:len(buf)] = buf
        return len(buf)


class _ObjectWriter:
    def __init__(self, client, bucket, key, mode, mpu_chunksize, kwargs):
        self.client = client
        self.bucket = bucket
        self.key = key
        self.mode = mode
        self._init_buf()
        self.mpu_chunksize = mpu_chunksize
        self.mpu_id = None
        self.parts = []

    def _init_buf(self):
        if 'b' in self.mode:
            self.buf = io.BytesIO()
        else:
            self.buf = io.StringIO()

    def flush(self):
        # A part must be more than 8 MiB in S3
        if len(self.buf.getvalue()) < 8 * 1024 * 1024:
            return
        self._flush()

    def _flush(self):
        # Send buffer as a part
        c = self.client
        b = self.bucket
        k = self.key

        if self.mpu_id is None:
            res = c.create_multipart_upload(Bucket=b, Key=k)
            self.mpu_id = res['UploadId']
            boto3.set_stream_logger()

        assert self.mpu_id is not None

        data = self.buf.getvalue()
        if 'b' in self.mode:
            md5 = hashlib.md5(data).hexdigest()
        else:
            md5 = hashlib.md5(data.encode()).hexdigest()
        num = len(self.parts) + 1

        res = c.upload_part(Body=data, Bucket=b, Key=k,
                            PartNumber=num,
                            UploadId=self.mpu_id,
                            ContentLength=len(data),
                            ContentMD5=md5)
        self.parts.append({'ETag': res['ETag'], 'PartNumber': num})

        self._init_buf()

    def write(self, buf):
        written = 0
        overflow = len(self.buf.getvalue()) + len(buf) - self.mpu_chunksize
        if overflow > 0:
            l = len(buf) - overflow
            written += self.buf.write(buf[:l])
            self.flush()
            buf = buf[l:]

        written += self.buf.write(buf)
        if len(self.buf.getvalue()) >= self.mpu_chunksize:
            self.flush()

        return written

    def close(self):
        # See:  https://boto3.amazonaws.com/v1/documentation/
        # api/latest/reference/services/s3.html#S3.Client.put_object
        if len(self.parts) == 0:
            self.client.put_object(Body=self.buf.getvalue(),
                                   Bucket=self.bucket,
                                   Key=self.key)
        else:
            self._flush()
            # DO: MPU
            c = self.client
            max_parts = len(self.parts) + 1
            res = c.list_parts(Bucket=self.bucket,
                               Key=self.key,
                               UploadId=self.mpu_id, MaxParts=max_parts)

            if res['IsTruncated']:
                next_part = res['NextPartNumberMarker']
                raise RuntimeError('Unexpectedly truncated: ' +
                                   'next={}/maxparts={}'.format(next_part,
                                                                max_parts))

            parts = [{'ETag': part['ETag'], 'PartNumber': part['PartNumber']}
                     for part in res.get('Parts', [])]
            parts = sorted(parts, key=lambda x: int(x['PartNumber']))
            assert self.parts == parts

            res = c.complete_multipart_upload(Bucket=self.bucket,
                                              Key=self.key,
                                              UploadId=self.mpu_id,
                                              MultipartUpload={'Parts': parts})
            # logger.info("Upload done.", res['Location'])

        self.buf = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        self.close()

    @property
    def closed(self):
        return self.buf is None

    def isatty(self):
        return False

    def readable(self):
        return False

    def seekable(self):
        return False

    def writable(self):
        return True


class S3(FS):
    '''S3 FileSystem wrapper

    Takes three arguments as well as enviroment variables for
    constructor. The priority is (1) see arguments, (2) see enviroment
    variables, (3) take boto3's default. Available arguments are:

    - ``aws_access_key_id``, ``AWS_ACCESS_KEY_ID``
    - ``aws_secret_access_key``, ``AWS_SECRET_ACCESS_KEY``
    - ``endpoint``, ``S3_ENDPOINT``

    It supports buffering when opening a file in binary read mode ("rb").
    When ``buffering`` is set to -1 (default), the buffer size will be
    the size of the file or ``pfio.v2.S3.DEFAULT_MAX_BUFFER_SIZE``,
    whichever smaller.
    ``buffering=0`` disables buffering, and ``buffering>0`` forcibly sets the
    specified value as the buffer size in bytes.
    '''

    def __init__(self, bucket, prefix=None,
                 endpoint=None, create_bucket=False,
                 aws_access_key_id=None,
                 aws_secret_access_key=None,
                 mpu_chunksize=32*1024*1024,
                 buffering=-1,
                 create=False,
                 reset_on_fork=False,
                 **_):
        super().__init__(reset_on_fork=reset_on_fork)
        self.bucket = bucket
        self.create_bucket = create_bucket
        if prefix is not None:
            self.cwd = prefix
        else:
            self.cwd = ''

        # In S3, create flag can be disregarded
        del create

        self.mpu_chunksize = mpu_chunksize
        self.buffering = buffering

        # boto3.set_stream_logger()

        # import botocore
        # botocore.session.Session().set_debug_logger()

        kwargs = {}

        # IF these arguments are not defined, the library
        # automatically retrieves from AWS_ACCESS_KEY_ID and
        # AWS_SECRET_ACCESS_KEY.
        self.aws_access_key_id = aws_access_key_id
        if aws_access_key_id is not None:
            kwargs['aws_access_key_id'] = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        if aws_secret_access_key is not None:
            kwargs['aws_secret_access_key'] = aws_secret_access_key

        # We won't expect any enviroment variable for S3 endpoints
        # supported by boto3. Instead, we take S3_ENDPOINT in case
        # argument ``endpoint`` is not given. Otherwise, it goes to
        # boto3's default by giving ``None``.
        #
        # See also:
        # https://github.com/boto/boto3/issues/1375
        # https://github.com/boto/boto3/pull/2746
        self.endpoint = endpoint
        if self.endpoint is None:
            self.endpoint = os.getenv('S3_ENDPOINT')
        if self.endpoint is not None:
            kwargs['endpoint_url'] = self.endpoint

        self.kwargs = kwargs
        self._connect()

    def _reset(self):
        self._connect()

    def _connect(self):
        # print('boto3.client options:', kwargs)
        self.client = boto3.client('s3', **self.kwargs)

        try:
            self.client.head_bucket(Bucket=self.bucket)
        except ClientError as e:
            if e.response['Error']['Code'] == '404' and self.create_bucket:
                res = self.client.create_bucket(Bucket=self.bucket)
                print("Bucket", self.bucket, "created:", res)
            else:
                raise e

    def __getstate__(self):
        state = self.__dict__.copy()
        state['client'] = None
        return state

    def __setstate__(self, state):
        self.__dict__ = state

    def open(self, path, mode='r', **kwargs):
        '''Opens an object accessor for read or write

        .. note:: Multi-part upload is not yet available.

        Arguments:
            path (str): relative path from basedir

            mode (str): open mode
        '''
        self._checkfork()
        if 'a' in mode:
            raise io.UnsupportedOperation('Append is not supported')
        if 'r' in mode and 'w' in mode:
            raise io.UnsupportedOperation('Read-write mode is not supported')

        path = os.path.join(self.cwd, path)
        path = _normalize_key(path)
        if 'r' in mode:
            obj = _ObjectReader(self.client, self.bucket, path, mode, kwargs)

            bs = self.buffering
            if bs < 0:
                bs = min(obj.content_length, DEFAULT_MAX_BUFFER_SIZE)

            if 'b' in mode:
                if self.buffering and bs != 0:
                    obj = io.BufferedReader(obj, buffer_size=bs)
            else:
                obj = io.TextIOWrapper(obj)
                if self.buffering:
                    # This is undocumented property; but resident at
                    # least since 2009 (the merge of io-c branch).
                    # We'll use it until the day of removal.
                    if bs == 0:
                        # empty file case: _CHUNK_SIZE must be positive
                        bs = DEFAULT_MAX_BUFFER_SIZE
                    obj._CHUNK_SIZE = bs

        elif 'w' in mode:
            obj = _ObjectWriter(self.client, self.bucket, path, mode,
                                self.mpu_chunksize, kwargs)
            if 'b' in mode:
                obj = io.BufferedWriter(obj)

        else:
            raise RuntimeError(f'Unknown option: {mode}')

        return obj

    def list(self, prefix: str = "", recursive=False):
        '''List all objects (and prefixes)

        Although there is not concept of directory in AWS S3 API,
        common prefixes shows up like directories.

        '''
        self._checkfork()
        key = os.path.join(self.cwd, prefix)
        key = _normalize_key(key)
        if key == '.':
            key = ''
        elif key != '' and not key.endswith('/'):
            key += '/'
        if '/../' in key or key.startswith('..'):
            raise ValueError('Invalid S3 key: {} as {}'.format(prefix, key))

        page_size = 1000
        paginator = self.client.get_paginator('list_objects_v2')
        paging_args = {
            'Bucket': self.bucket, 'Prefix': key,
            'PaginationConfig': {'PageSize': page_size}
        }
        if not recursive:
            paging_args['Delimiter'] = '/'

        iterator = paginator.paginate(**paging_args)
        for res in iterator:
            # print(res)
            for common_prefix in res.get('CommonPrefixes', []):
                yield common_prefix['Prefix'][len(key):]
            for content in res.get('Contents', []):
                yield content['Key'][len(key):]

    def stat(self, path):
        '''Imitate FileStat with S3 Object metadata

        '''
        self._checkfork()
        key = os.path.join(self.cwd, path)
        key = _normalize_key(key)
        try:
            res = self.client.head_object(Bucket=self.bucket,
                                          Key=key)
            if res.get('DeleteMarker'):
                raise FileNotFoundError()

            return S3ObjectStat(key, res)
        except ClientError as e:
            if e.response['Error']['Code'] == '404':
                if self.isdir(path):
                    return S3PrefixStat(key)
                raise FileNotFoundError()
            else:
                raise e

    def isdir(self, file_path: str):
        '''Imitate isdir by handling common prefix ending with "/" as directory

        AWS S3 does not have concept of directory tree, but this class
        imitates other file systems to increase compatibility.
        '''
        self._checkfork()
        key = _normalize_key(os.path.join(self.cwd, file_path))
        if key == '.':
            key = ''
        elif key.endswith('/'):
            key = key[:-1]
        if '/../' in key or key.startswith('..'):
            raise ValueError('Invalid S3 key: {} as {}'.format(file_path, key))

        if len(key) == 0:
            return True

        res = self.client.list_objects_v2(
            Bucket=self.bucket,
            Prefix=key,
            Delimiter="/",
            MaxKeys=1,
        )
        for common_prefix in res.get('CommonPrefixes', []):
            if common_prefix['Prefix'] == key + "/":
                return True
        return False

    def mkdir(self, file_path: str, mode=0o777, *args, dir_fd=None):
        '''Does nothing

        .. note:: AWS S3 does not have concept of directory tree; what
           this function (and ``makedirs()``) should do
           and return? To be strict, it would be straightforward to
           raise ``io.UnsupportedOperation`` exception. But it just
           breaks users' applications that except quasi-compatible
           behaviour. Thus, imitating other file systems, like
           returning ``None`` would be nicer.
        '''
        # raise io.UnsupportedOperation("S3 doesn't have directory")
        pass

    def makedirs(self, file_path: str, mode=0o777, exist_ok=False):
        '''Does nothing

        .. note:: see discussion in ``mkdir()``.
        '''
        # raise io.UnsupportedOperation("S3 doesn't have directory")
        pass

    def exists(self, file_path: str):
        '''Returns the existence of objects

        For common prefixes, it does nothing. See discussion in ``isdir()``.
        '''
        self._checkfork()
        try:
            key = os.path.join(self.cwd, file_path)
            key = _normalize_key(key)
            res = self.client.head_object(Bucket=self.bucket,
                                          Key=key)
            return not res.get('DeleteMarker')
        except ClientError as e:
            if e.response['Error']['Code'] == '404':
                if self.isdir(file_path):
                    return True
                return False
            else:
                raise e

    def rename(self, src, dst):
        '''Copies & removes the object

        Source and destination must be in the same bucket for
        ``pfio``, although AWS S3 supports inter-bucket copying.

        '''
        self._checkfork()
        source = {
            'Bucket': self.bucket,
            'Key': _normalize_key(os.path.join(self.cwd, src)),
        }
        dst = os.path.join(self.cwd, dst)
        dst = _normalize_key(dst)
        res = self.client.copy_object(Bucket=self.bucket,
                                      CopySource=source,
                                      Key=dst)
        if not res.get('CopyObjectResult'):
            # copy failed
            return
        return self.remove(src)

    def remove(self, file_path: str, recursive=False):
        '''Removes an object

        It raises a FileNotFoundError when the specified file doesn't exist.
        '''
        if recursive:
            raise io.UnsupportedOperation("Recursive delete not supported")

        if not self.exists(file_path):
            msg = "No such S3 object: '{}'".format(file_path)
            raise FileNotFoundError(msg)

        self._checkfork()
        key = os.path.join(self.cwd, file_path)
        key = _normalize_key(key)
        return self.client.delete_object(Bucket=self.bucket,
                                         Key=key)



import io
import mmap
import os
import pickle
from struct import calcsize, unpack

from pfio import cache
from pfio.cache.file_cache import _default_cache_path


class ReadOnlyFileCache(cache.Cache):
    '''Read only cache system with local filesystem

    Example::

      c = ReadOnlyFileCache(65536, dir="path/to/cache")
      c.preload("filename")

    Arguments:
        length (int): Length of the cache array.

        do_pickle (bool):
            Do automatic pickle and unpickle inside the cache.

        verbose (bool):
            Print detailed logs of the cache.

    '''

    def __init__(self, length, do_pickle=False, dir=None, verbose=False):
        self.length = length
        self.do_pickle = do_pickle
        if self.length <= 0 or (2 ** 64) <= self.length:
            raise ValueError("length has to be between 0 and 2^64")

        if dir is None:
            self.dir = _default_cache_path()
        else:
            self.dir = dir
        cache.file_cache._check_local(self.dir)

        self.buflen = calcsize('Qq')
        assert self.buflen == 16

        self.closed = True
        self.verbose = verbose

    def __len__(self):
        return self.length

    @property
    def frozen(self):
        return True

    @property
    def multiprocess_safe(self):
        return True

    @property
    def multithread_safe(self):
        return True

    def get(self, i):
        if self.closed:
            return
        data = self._get(i)
        if self.do_pickle and data:
            data = pickle.loads(data)
        return data

    def _get(self, i):
        if i < 0 or self.length <= i:
            raise IndexError("index {} out of range ([0, {}])"
                             .format(i, self.length - 1))

        (o, l) = self.offset_list[i]
        return self.mreg[o:(o+l)]

    def put(self, i, data):
        raise io.UnsupportedOperation('read only cache')

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self):
        if not self.closed:
            self.mreg.close()
            self.fp.close()
            self.closed = True

    def preload(self, name):
        '''Load the cache saved by ``preserve()``

        ``cache_path`` is the path to the persistent file. To use cache
        in ``multiprocessing`` environment, call this method at every
        forked process, except the process that called ``preserve()``.
        After the preload, no data can be added to the cache.

        When it succeeds, it returns ``True``.
        If there is no cache file with the specified name in
        the cache directory, it will do nothing but return ``False``.

        Returns:
            bool: Returns True if succeed.

        .. note:: This feature is experimental.

        '''
        filename = os.path.join(self.dir, name)
        self.fp = open(filename, 'rb')

        self.offset_list = []
        for i in range(self.length):
            offset = self.buflen * i
            buf = os.pread(self.fp.fileno(), self.buflen, offset)
            (o, l) = unpack('Qq', buf)
            if l < 0 or o < 0:
                raise RuntimeError("Incomplete cache file")

            self.offset_list.append((o, l))

        assert self.length == len(self.offset_list)

        # mmaped region
        self.mreg = mmap.mmap(self.fp.fileno(), 0, prot=mmap.PROT_READ)
        self.closed = False
        return True

    def preserve(self, name, overwrite=False):
        raise io.UnsupportedOperation('read only')



import io
import logging
import os
import zipfile
from datetime import datetime

from pfio.cache.sparse_file import CachedWrapper, MPCachedWrapper

from .fs import FS, FileStat

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


class ZipFileStat(FileStat):
    """Detailed information of a file in a Zip

    Attributes:
        filename (str): Derived from `~FileStat`.
        orig_filename (str): ``ZipFile.orig_filename``.
        comment (str): ``ZipFile.comment``.
        last_modifled (float): Derived from `~FileStat`.
            No sub-second precision.
        mode (int): Derived from `~FileStat`.
        size (int): Derived from `~FileStat`.
        create_system (int): ``ZipFile.create_system``.
        create_version (int): ``ZipFile.create_version``.
        extract_version (int): ``ZipFile.extract_version``.
        flag_bits (int): ``ZipFile.flag_bits``.
        volume (int): ``ZipFile.volume``.
        internal_attr (int): ``ZipFile.internal_attr``.
        external_attr (int): ``ZipFile.external_attr``.
        header_offset (int): ``ZipFile.header_offset``.
        compress_size (int): ``ZipFile.compress_size``.
        compress_type (int): ``ZipFile.compress_type``.
        CRC (int): ``ZipFile.CRC``.
    """

    def __init__(self, zip_info):
        self.last_modified = float(datetime(*zip_info.date_time).timestamp())
        # https://github.com/python/cpython/blob/3.8/Lib/zipfile.py#L392
        self.mode = zip_info.external_attr >> 16
        self.size = zip_info.file_size

        for k in ('filename', 'orig_filename', 'comment', 'create_system',
                  'create_version', 'extract_version', 'flag_bits',
                  'volume', 'internal_attr', 'external_attr', 'CRC',
                  'header_offset', 'compress_size', 'compress_type'):
            setattr(self, k, getattr(zip_info, k))


class Zip(FS):
    _readonly = True
    '''
            local_cache (bool): use sparse file cache for opening ZIP file

            local_cachedir (dir): local path to store sparse file cache
    '''

    def __init__(self, backend, file_path, mode='r', create=False,
                 local_cache=False, local_cachedir=None,
                 reset_on_fork=False, **kwargs):
        super().__init__(reset_on_fork=reset_on_fork)
        self.backend = backend
        self.file_path = file_path
        self.mode = mode
        self.kwargs = kwargs

        if create:
            raise ValueError("create option is not supported")

        if 'r' in mode and 'w' in mode:
            raise io.UnsupportedOperation('Read-write mode is not supported')

        if 'w' in mode:
            self._readonly = False

        self.local_cache = local_cache
        self.local_cachedir = local_cachedir
        self.local_cachefile = None
        self.local_indexfile = None

        self._reset()

    def _reset(self):
        buffering = self.kwargs.get('buffering', 16*1024*1024)

        if self.local_cache:
            # Don't use io.BufferedReader for sparse file cache
            self.kwargs['buffering'] = 0

        obj = self.backend.open(self.file_path,
                                self.mode + 'b',
                                **self.kwargs)

        stat = self.backend.stat(self.file_path)

        # Use sparse file cache: Optimization for a remote object
        # store system e.g. AWS S3 or HDFS
        if self.local_cache:
            self.kwargs['buffering'] = buffering

            if self.reset_on_fork:
                obj = MPCachedWrapper(obj, stat.size, self.local_cachedir,
                                      local_cachefile=self.local_cachefile,
                                      local_indexfile=self.local_indexfile,
                                      close_on_close=True,
                                      multithread_safe=True)

                # Update local cachefile in case of being forked
                self.local_cachefile = obj.local_cachefile
                self.local_indexfile = obj.local_indexfile
            else:
                obj = CachedWrapper(obj, stat.size, self.local_cachedir,
                                    close_on_close=True, multithread_safe=True)

                # Default 16MB buffer size
            if buffering > 0:
                obj = io.BufferedReader(obj, buffer_size=buffering)

        self.fileobj = obj

        assert self.fileobj is not None
        self.zipobj = zipfile.ZipFile(self.fileobj, self.mode)

    def open(self, file_path, mode='r',
             buffering=-1, encoding=None, errors=None,
             newline=None, closefd=True, opener=None):
        self._checkfork()

        file_path = os.path.join(self.cwd, os.path.normpath(file_path))
        fp = self.zipobj.open(file_path, mode.replace('b', ''))

        if 'b' not in mode:
            fp = io.TextIOWrapper(fp, encoding, errors, newline)

        return fp

    def subfs(self, path):
        # TODO
        raise NotImplementedError()

    def close(self):
        self._checkfork()
        self.zipobj.close()
        self.fileobj.close()

    def stat(self, path):
        self._checkfork()
        path = os.path.join(self.cwd, os.path.normpath(path))
        if path in self.zipobj.namelist():
            actual_path = path
        elif (not path.endswith('/')
              and path + '/' in self.zipobj.namelist()):
            # handles cases when path is a directory but without trailing slash
            # see issue $67
            actual_path = path + '/'
        else:
            raise FileNotFoundError(
                "{} is not found".format(path))

        return ZipFileStat(self.zipobj.getinfo(actual_path))

    def list(self, path_or_prefix: str = "", recursive=False):
        self._checkfork()

        if path_or_prefix:
            path_or_prefix = os.path.join(self.cwd,
                                          os.path.normpath(path_or_prefix))
            # cannot move beyond root
            given_dir_list = path_or_prefix.split('/')
            if ("." in given_dir_list or ".." in given_dir_list
                    or {""} == set(given_dir_list)):
                given_dir_list = []
                path_or_prefix = ""
        else:
            given_dir_list = []

        if path_or_prefix:
            if self.exists(path_or_prefix) and not self.isdir(path_or_prefix):
                raise NotADirectoryError(
                    "{} is not a directory".format(path_or_prefix))
            elif not any(name.startswith(path_or_prefix + "/")
                         for name in self.zipobj.namelist()):
                # check if directories are NOT included in the zip
                # such kind of zip can be made with "zip -D"
                raise FileNotFoundError(
                    "{} is not found".format(path_or_prefix))

        if recursive:
            for name in self.zipobj.namelist():
                if name.startswith(path_or_prefix):
                    name = name[len(path_or_prefix):].strip("/")
                    if name:
                        yield name
        else:
            _list = set()
            for name in self.zipobj.namelist():
                return_file_name = None
                current_dir_list = os.path.normpath(name).split('/')
                if not given_dir_list:
                    # if path_or_prefix is not given
                    return_file_name = current_dir_list[0]
                else:
                    if (current_dir_list
                            and len(current_dir_list) > len(given_dir_list)
                            and current_dir_list[:len(given_dir_list)] ==
                            given_dir_list):
                        return_file_name = current_dir_list[
                            len(given_dir_list):][0]

                if (return_file_name is not None
                        and return_file_name not in _list):
                    _list.add(return_file_name)
                    yield return_file_name

    def isdir(self, file_path: str):
        self._checkfork()
        file_path = os.path.join(self.cwd, file_path)
        if self.exists(file_path):
            return self.stat(file_path).isdir()
        else:
            file_path = os.path.normpath(file_path)
            # check if directories are NOT included in the zip
            if any(name.startswith(file_path + "/")
                   for name in self.zipobj.namelist()):
                return True

            return False

    def mkdir(self, file_path: str, mode=0o777, *args, dir_fd=None):
        raise io.UnsupportedOperation("zip does not support mkdir")

    def makedirs(self, file_path: str, mode=0o777, exist_ok=False):
        raise io.UnsupportedOperation("zip does not support makedirs")

    def exists(self, file_path: str):
        self._checkfork()
        file_path = os.path.join(self.cwd, os.path.normpath(file_path))
        namelist = self.zipobj.namelist()
        return (file_path in namelist
                or file_path + "/" in namelist)

    def rename(self, *args):
        raise io.UnsupportedOperation

    def remove(self, file_path, recursive=False):
        raise io.UnsupportedOperation



import getpass
import io
import logging
import multiprocessing
import os
import re
import subprocess
import warnings
from xml.etree import ElementTree

import pyarrow
from pyarrow.fs import FileSelector, FileType, HadoopFileSystem

from .fs import FS, FileStat

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


def _parse_principal_name_from_klist(output):
    output_array = output.split('\n')
    if len(output_array) < 2:
        return None

    principle_str = output_array[1]
    klist_principal_pattern = re.compile(
        r'Default principal: (?P<username>.+)@(?P<service>.+)')
    ret = klist_principal_pattern.match(principle_str)
    if ret:
        pattern_dict = ret.groupdict()
        return pattern_dict['username']
    else:
        return None


def _parse_principal_name_from_keytab(output):
    output_array = output.split('\n')
    if len(output_array) < 4:
        return None

    principle_str = output_array[3]
    keytab_principle_pattern = re.compile(
        r'\s+\d+ (?P<username>.+)@(?P<service>.+)')
    ret = keytab_principle_pattern.match(principle_str)
    if ret:
        pattern_dict = ret.groupdict()
        return pattern_dict['username']
    else:
        return None


def _get_principal_name_from_keytab():
    output = _run_klist(use_keytab=True)
    if output is None:
        return None

    return _parse_principal_name_from_keytab(output.decode('utf-8'))


def _get_principal_name_from_klist():
    output = _run_klist()
    if output is None:
        return None
    return _parse_principal_name_from_klist(output.decode('utf-8'))


def _run_klist(use_keytab=False):
    try:
        command = ['klist']
        if use_keytab:
            command += ['-k']
        pipe = subprocess.Popen(command, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        out, err = pipe.communicate()
        if out == b'' and err != b'':
            return None
        else:
            return out
    except OSError:
        # klist is not found
        return None


class HdfsFileStat(FileStat):
    """Detailed information of a file in HDFS

    Attributes:
        filename (str): Derived from `~FileStat`.
        last_modifled (float): Derived from `~FileStat`.
            No sub-second precision.
        last_accessed (float): UNIX timestamp of last access time.
            No sub-second precision.
        mode (int): Derived from `~FileStat`.
        size (int): Derived from `~FileStat`.
    """

    def __init__(self, info):
        self._info = info

        mode = 0
        if info.type == FileType.File:
            mode |= 0o100000
        elif info.type == FileType.Directory:
            mode |= 0o40000

        self.filename = info.base_name
        self.mode = mode
        self.last_modified = info.mtime.timestamp()
        # no atime supported by PyArrow new API
        self.last_accessed = info.mtime.timestamp()
        self.size = info.size


def _ensure_arrow_envs(hadoop_home):
    if os.getenv("ARROW_LIBHDFS_DIR") is None:
        arrow_dir = os.path.join(hadoop_home, "lib")
        libhdfs = os.path.join(arrow_dir, "libhdfs.so")

        if os.path.exists(libhdfs):
            os.environ["ARROW_LIBHDFS_DIR"] = arrow_dir
        else:
            arrow_dir = os.path.join(hadoop_home, "lib/native")
            libhdfs = os.path.join(arrow_dir, "libhdfs.so")
            if os.path.exists(libhdfs):
                os.environ["ARROW_LIBHDFS_DIR"] = arrow_dir
            else:
                msg = "No libhdfs.so found from $HADOOP_HOME: {}".format(
                    hadoop_home)
                raise RuntimeError(msg)


def _create_fs():
    confdir = os.getenv('HADOOP_CONF_DIR', '/etc/hadoop/conf')
    conffile = os.path.join(confdir, 'hdfs-site.xml')
    root = ElementTree.parse(conffile)

    # Valid envs required. Typically, /opt/cloudera/parcels/CDH/lib
    if os.getenv("HADOOP_HOME"):
        hadoop_home = os.getenv("HADOOP_HOME")
        _ensure_arrow_envs(hadoop_home)

    if os.getenv("CLASSPATH") is None:
        # TODO(kuenishi): CLASSPATH=`$HADOOP_HOME/bin/hdfs classpath --glob`
        cmd = ["hdfs", "classpath", "--glob"]
        cp = subprocess.run(cmd, stdout=subprocess.PIPE)
        cp.check_returncode()
        os.environ["CLASSPATH"] = cp.stdout.decode()

    assert os.getenv("ARROW_LIBHDFS_DIR"), "LIBHDFS not found"
    assert os.getenv("CLASSPATH"), "CLASSPATH not defined"

    configs = {}
    for e in root.findall('./property'):
        name = None
        for c in e:
            if c.tag == 'name':
                name = c.text
            elif c.tag == 'value':
                value = c.text
        if name:
            configs[name] = value

    for nameservice in configs.get('dfs.nameservices', '').split(','):

        # TODO(kuenishi): We don't have such use case where we switch
        # amont multiple name services from single HADOOP_CONF_DIR
        # conf. Thus we ignore fs.defaultFS and just take the very
        # first name service that appeared in hdfs-site.xml.
        return HadoopFileSystem(nameservice, 0)

    else:
        RuntimeError("No nameservice found.")


class Hdfs(FS):
    '''Hadoop FileSystem wrapper

    To use HDFS, PFIO requires ``$HADOOP_HOME`` predefined before
    initialization. If it is not defined, ``ARROW_LIBHDFS_DIR`` must
    be defined instead. ``$CLASSPATH`` will be needed in case ``hdfs``
    command is not available from ``$PATH``.


    .. warning:: It is strongly discouraged to use :class:`Hdfs` under
          multiprocessing. Setting ``reset_on_fork=False`` is
          recommended for HDFS in order not to shoot yourself in the
          foot, because forking a JVM instance that has a different
          memory model may cause unexpected behavior. If you do *need*
          forking, for example, PyTorch DataLoader with multiple
          workers for performance, it is strongly recommended not to
          instantiate :class:`Hdfs` before forking. Details are
          described in PFIO issue #123.  Simple workaround is to set
          multiprocessing start method as ``'forkserver'`` and start
          the very first child process before everything.

          .. code-block::

             import multiprocessing
             multiprocessing.set_start_method('forkserver')
             p = multiprocessing.Process()
             p.start()
             p.join()


          See: https://github.com/pfnet/pfio/issues/123

    .. note:: With environment variable
          ``KRB5_KTNAME=path/to/your.keytab`` set, ``hdfs``
          handler automatically starts automatic and periodical
          updating Kerberos ticket using `krbticket
          <https://pypi.org/project/krbticket/>`_ . The update
          frequency is every 10 minutes by default.
    .. note::
          Only the username in the first entry in The
          keytab will be used to update the Kerberos ticket.

    '''

    def __init__(self, cwd=None, create=False, reset_on_fork=False, **_):
        super().__init__(reset_on_fork=reset_on_fork)
        self._fs = _create_fs()
        assert self._fs is not None
        self.username = self._get_principal_name()

        self.cwd = os.path.join('/user', self.username)

        if cwd is not None:
            if cwd.startswith('/'):
                self.cwd = cwd
            else:
                self.cwd = os.path.join(self.cwd, cwd)

        if not self.isdir(''):
            if create:
                # Since this process (isdir -> makedirs) is not atomic,
                # makedirs can conflict in case of a parallel workload.
                self.makedirs('', exist_ok=True)
            else:
                raise ValueError('{} must be a directory'.format(self.cwd))

        if reset_on_fork and\
           multiprocessing.get_start_method() != 'forkserver':
            # See https://github.com/pfnet/pfio/pull/123 for detail
            warnings.warn('Non-forkserver start method under HDFS detected.')

    def _reset(self):
        self._fs = _create_fs()

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_fs'] = None
        return state

    def __setstate__(self, state):
        self.__dict__ = state

    def _get_principal_name(self):
        # get the default principal name from `klist` cache
        principal_name = _get_principal_name_from_klist()

        if principal_name is not None:
            return principal_name

        # try getting principal name from keytab
        principal_name = _get_principal_name_from_keytab()
        if principal_name is not None:
            return principal_name

        # in case every thing, use the login username instead
        return self._get_login_username()

    def _get_login_username(self):
        return getpass.getuser()

    def open(self, file_path, mode='rb',
             buffering=-1, encoding=None, errors=None,
             newline=None, closefd=True, opener=None):
        self._checkfork()
        path = os.path.join(self.cwd, file_path)

        try:
            if 'r' in mode:
                file_obj = self._fs.open_input_file(path)
            else:
                file_obj = self._fs.open_output_stream(path)
        except pyarrow.lib.ArrowIOError as e:
            raise IOError("open file error :{}".format(str(e)))

        return self._wrap_file_obj(file_obj, mode, encoding, errors, newline)

    def _wrap_file_obj(self, file_obj, mode, encoding, errors, newline):
        if 'b' not in mode:
            return io.TextIOWrapper(file_obj, encoding, errors, newline)
        elif 'r' in mode:
            # Wrap file_obj with io.BufferedReader for ``peek()``, to
            # significiantly improve unpickle performance.
            return io.BufferedReader(file_obj)
        elif 'w' in mode:
            return io.BufferedWriter(file_obj)
        else:
            raise ValueError("invalid option")

    def subfs(self, rel_path):
        return Hdfs(os.path.join(self.cwd, rel_path))

    def close(self):
        pass

    def list(self, path: str = "", recursive=False):
        self._checkfork()

        if not self.isdir(path):
            raise NotADirectoryError(path)

        path = os.path.join(self.cwd, path)
        norm_path = self._fs.normalize_path(path).rstrip('/')

        infos = self._fs.get_file_info(FileSelector(path, recursive=recursive))
        for file_info in infos:
            yield file_info.path[len(norm_path)+1:]

    def stat(self, path):
        self._checkfork()
        path = os.path.join(self.cwd, path)
        info = self._fs.get_file_info(path)
        if info.type == FileType.NotFound:
            raise FileNotFoundError()
        else:
            return HdfsFileStat(info)

    def isdir(self, path: str):
        self._checkfork()
        path = os.path.join(self.cwd, path)
        info = self._fs.get_file_info(path)
        return info.type == FileType.Directory

    def mkdir(self, path: str, *args, dir_fd=None):
        self._checkfork()
        path = os.path.join(self.cwd, path)
        return self._fs.create_dir(path)

    def makedirs(self, path: str, mode=0o777, exist_ok=False):
        self._checkfork()
        if self.exists(path) and self.isdir(path) and not exist_ok:
            raise NotADirectoryError()
        path = os.path.join(self.cwd, path)
        return self._fs.create_dir(path, recursive=True)

    def exists(self, path: str):
        path = os.path.join(self.cwd, path)
        info = self._fs.get_file_info(path)
        return info.type != FileType.NotFound

    def rename(self, src, dst):
        self._checkfork()
        s = os.path.join(self.cwd, src)
        d = os.path.join(self.cwd, dst)
        return self._fs.move(s, d)

    def remove(self, path, recursive=False):
        delpath = os.path.join(self.cwd, path)
        if self.isdir(path):
            if recursive:
                self._fs.delete_dir(delpath)
            elif list(self.list(path)):
                raise RuntimeError("Directory not empty: {}".format(delpath))
            else:
                self._fs.delete_dir(delpath)
        else:
            self._fs.delete_file(delpath)



'''
fs.FS> interface
implementations:
open_fs(URI, container=None|zip) => Local/HDFS/S3/Zip, etc
- Local
  - subfs() -> Local
  - open_zip() -> Zip
  - open() -> FileObject
- HDFS
  - subfs() -> HDFS
  - open_zip() -> Zip
  - open() -> FileObject
- Zip
  - subfs() -> Zip
  - open_zip() -> Zip
  - open() -> FileObject
- S3 (TBD)
- GS (TBD)

For example of globally switching backend file systems::

  from pfio.v2 import local as pfio

Or::

  from pfio.v2 import Hdfs
  pfio = Hdfs()

'''
from .fs import from_url, lazify, open_url  # NOQA
from .hdfs import Hdfs, HdfsFileStat  # NOQA
from .local import Local, LocalFileStat  # NOQA
from .pathlib import Path  # NOQA
from .s3 import S3  # NOQA
from .zip import Zip, ZipFileStat  # NOQA

local = Local()



import io
import os
import shutil

from .fs import FS, FileStat


class LocalFileStat(FileStat):
    """Detailed information of a POSIX file

    The information of file/directory is obtained through the `os.stat`.

    Attributes:
        filename (str): Derived from `~FileStat`.
        last_modified (float): Derived from `~FileStat`.
            ``os.stat_result.st_mtime``.
        last_accessed (float): ``os.stat_result.st_atime``.
        created (float): ``os.stat_result.st_ctime``.
        last_modified_ns (int): ``os.stat_result.st_mtime_ns``.
        last_accessed_ns (int): ``os.stat_result.st_atime_ns``.
        created_ns (float): ``os.stat_result.st_ctime``.
        mode (int): Derived from `~FileStat`. ``os.stat_result.st_mode``.
        size (int): Derived from `~FileStat`. ``os.stat_result.st_size``.
        owner (int): UID of owner in integer.
        group (int): GID of the file in integer.
        inode (int): ``os.stat_result.st_ino``.
        device (int): ``os.stat_result.st_dev``.
        nlink (int): ``os.stat_result.st_nlink``.
    """

    def __init__(self, _stat, filename):
        keys = (('last_modified', 'st_mtime'),
                ('last_accessed', 'st_atime'),
                ('last_modified_ns', 'st_mtime_ns'),
                ('last_accessed_ns', 'st_atime_ns'),
                ('created', 'st_ctime'), ('created_ns', 'st_ctime_ns'),
                ('mode', 'st_mode'), ('size', 'st_size'), ('uid', 'st_uid'),
                ('gid', 'st_gid'), ('ino', 'st_ino'), ('dev', 'st_dev'),
                ('nlink', 'st_nlink'))
        for k, ksrc in keys:
            setattr(self, k, getattr(_stat, ksrc))
        self.filename = filename


class Local(FS):
    def __init__(self, cwd=None, create=False, reset_on_fork=False, **_):
        super().__init__(reset_on_fork=reset_on_fork)

        if cwd is None:
            self._cwd = ''
        else:
            self._cwd = cwd

        if not self.isdir(''):
            if create:
                # Since this process (isdir -> makedirs) is not atomic,
                # makedirs can conflict in case of a parallel workload.
                os.makedirs(self._cwd, exist_ok=True)
            else:
                raise ValueError('{} must be a directory'.format(self._cwd))

    @property
    def cwd(self):
        if self._cwd:
            return self._cwd

        return os.getcwd()

    def _reset(self):
        pass

    def open(self, file_path, mode='r',
             buffering=-1, encoding=None, errors=None,
             newline=None, closefd=True, opener=None):

        path = os.path.join(self.cwd, file_path)
        return io.open(path, mode,
                       buffering, encoding, errors,
                       newline, closefd, opener)

    def list(self, path_or_prefix: str = '', recursive=False):
        path_or_prefix = os.path.join(self.cwd, path_or_prefix)

        if recursive:
            path_or_prefix = path_or_prefix.rstrip("/")
            # plus 1 to include the trailing slash
            prefix_end_index = len(path_or_prefix) + 1
            yield from self._recursive_list(prefix_end_index, path_or_prefix)
        else:
            for e in os.scandir(path_or_prefix):
                # ls -F
                if e.is_dir():
                    yield e.name + '/'
                else:
                    yield e.name

    def _recursive_list(self, prefix_end_index: int, path: str):
        for file in os.scandir(path):
            # ls -F
            if file.is_dir():
                yield file.path[prefix_end_index:] + '/'
                yield from self._recursive_list(prefix_end_index,
                                                file.path)
            else:
                yield file.path[prefix_end_index:]

    def stat(self, path):
        path = os.path.join(self.cwd, path)
        return LocalFileStat(os.stat(path), path)

    def isdir(self, path: str):
        path = os.path.join(self.cwd, path)
        return os.path.isdir(path)

    def mkdir(self, file_path: str, mode=0o777, *args, dir_fd=None):
        path = os.path.join(self.cwd, file_path)
        return os.mkdir(path, mode, *args, dir_fd=None)

    def makedirs(self, file_path: str, mode=0o777, exist_ok=False):
        path = os.path.join(self.cwd, file_path)
        return os.makedirs(path, mode, exist_ok)

    def exists(self, file_path: str):
        path = os.path.join(self.cwd, file_path)
        return os.path.exists(path)

    def rename(self, src, dst):
        s = os.path.join(self.cwd, src)
        d = os.path.join(self.cwd, dst)
        return os.rename(s, d)

    def remove(self, file_path: str, recursive=False):
        file_path = os.path.join(self.cwd, file_path)
        if recursive:
            return shutil.rmtree(file_path)
        if os.path.isdir(file_path):
            return os.rmdir(file_path)

        return os.remove(file_path)



'''
A cache system for remote file system that stores cache as mmap'ed file locally

Of course, it's read only
'''

import fcntl
import io
import os
import shutil
import struct
import tempfile
from contextlib import contextmanager
from dataclasses import dataclass
from types import TracebackType
from typing import Optional

from .file_cache import DummyLock, RWLock


@dataclass(frozen=True)
class _Range:
    start: int
    length: int
    cached: bool = False

    @staticmethod
    def size():
        return struct.calcsize('QLL')

    def overlap(self, rhs) -> bool:
        return (self.start - rhs.right) * (self.right - rhs.start) < 0

    @property
    def right(self):
        return self.start + self.length

    def includes(self, rhs) -> bool:
        return (self.start <= rhs.start) and (rhs.right <= self.right)

    def merge(self, rhs):
        assert self.overlap(rhs)
        assert self.cached == rhs.cached
        return _Range(min(self.start, rhs.start), max(self.right, rhs.right),
                      cached=self.cached)

    def pack(self):
        # Total 16 bytes
        # Q: unsigned long long: 8 bytes
        # L: unsigned long: 4 bytes
        # Upper bytes of cache is reserved for some other usage, maybe
        return struct.pack('QLL', self.start, self.cached, self.length)

    @staticmethod
    def unpack(buf):
        (s, c, l) = struct.unpack('QLL', buf)
        return _Range(start=s, cached=bool(c), length=l)


class _CachedWrapperBase:
    '''A transparent local cache for remote files base class

    TODO: add document here
    '''

    def __init__(self, fileobj, size, cachedir=None, close_on_close=False,
                 cache_size_limit=None, multithread_safe=False):
        self.fileobj = fileobj
        self.cachedir = cachedir
        self.cache_size_limit = cache_size_limit

        self.multithread_safe = multithread_safe
        if self.multithread_safe:
            self.lock = RWLock()
        else:
            self.lock = DummyLock()

        self.pos = 0
        self.size = size
        assert size > 0
        if cachedir is None:
            basedir = os.getenv('XDG_CACHE_HOME')
            if basedir is None:
                basedir = os.path.join(os.getenv('HOME'), ".cache")
            self.cachedir = os.path.join(basedir, "pfio")
        os.makedirs(self.cachedir, exist_ok=True)
        self.cachefp = tempfile.NamedTemporaryFile(delete=True,
                                                   dir=self.cachedir)

        # self.cachefp = open('cache.file', 'rwb')
        # self.cachefp = os.open('cache.file', os.O_RDWR|os.O_TRUNC)
        # TODO: make this tree if the size gets too long for O(n) scan
        self.ranges = [_Range(0, size)]
        self._closed = False
        self._frozen = False
        self.close_on_close = close_on_close

    def close(self):
        with self.lock.wrlock():
            if not self._closed:
                self._closed = True
                self.cachefp.close()
                if self.close_on_close:
                    self.fileobj.close()

    def preserve(self, dest):
        with self.lock.wrlock():
            # Hard link and save them
            try:
                os.link(self.cachefp.name, dest)
                self.cachefp.close()
            except OSError:
                # TODO: check errno to make sure handling the
                # 'different-drive' error.
                shutil.copyfile(self.cachefp.name, dest)

            self.cachefp = open(dest, 'rb')
            self._frozen = True

    def readline(self):
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Optional[BaseException],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        self.close()

    def flush(self):
        pass

    @property
    def closed(self):
        return self._closed

    def isatty(self):
        return False

    def readable(self):
        return True

    def seekable(self):
        return True

    def tell(self):
        return self.pos

    def truncate(self, size=None):
        raise io.UnsupportedOperation('truncate')

    def seek(self, pos, whence=io.SEEK_SET):
        with self.lock.wrlock():
            return self._seek(pos, whence)

    def _seek(self, pos, whence):
        # print(dir(self.fileobj))
        if whence in [0, io.SEEK_SET]:
            if pos < 0:
                raise OSError(22, "[Errno 22] Invalid argument")
        elif whence in [1, io.SEEK_CUR]:
            pos += self.pos
        elif whence in [2, io.SEEK_END]:
            pos += self.size
        else:
            raise ValueError('Wrong whence value: {}'.format(whence))

        if pos < 0:
            raise OSError(22, "[Errno 22] Invalid argument")
        self.pos = pos
        self.fileobj.seek(self.pos, io.SEEK_SET)
        return self.pos

    def writable(self):
        return False

    def write(self, data):
        raise io.UnsupportedOperation('not writable')

    def readall(self):
        return self.read(-1)

    def readinto(self, b):
        buf = self.read(len(b))
        b[:len(buf)] = buf
        return len(buf)

    def _read_all_cache(self):
        for r in self.ranges:
            if r.cached:
                data = os.pread(self.cachefp.fileno(), r.length, r.start)
                # print(r1, 'includes', r0, 'len(r0)?=', len(data))
                yield data, r
            else:
                yield None, r


class DynamicCachedWrapper(_CachedWrapperBase):
    '''Dynamic-page based local cache wrapper
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def read(self, size=-1) -> bytes:
        if size < 0:
            size = self.size - self.pos

        buf = bytearray(size)
        offset = 0
        with self.lock.wrlock():
            if self._closed:
                raise RuntimeError("closed")

            # TODO: unnecessary copy; optimize with os.readv?
            for data in self._read(size):
                buf[offset:offset+len(data)] = data
                offset += len(data)

            self.pos += len(buf)
            self.pos %= self.size
            if self.pos != self.fileobj.tell():
                self.fileobj.seek(self.pos, io.SEEK_SET)
            return bytes(buf)

    def _read(self, size):
        new_ranges = []
        streak = []
        for data, r in self._read2(size):
            if r.length > 0:
                if r.cached:
                    streak.append(r)
                else:
                    if streak:
                        start = streak[0].start
                        length = sum(s.length for s in streak)
                        new_ranges.append(_Range(start, length, cached=True))
                        streak = []
                    new_ranges.append(r)

            if data is not None:
                yield data

        if streak:
            start = streak[0].start
            length = sum(s.length for s in streak)
            new_ranges.append(_Range(start, length, cached=True))

        self.ranges = new_ranges

    def _read2(self, size):

        r0 = _Range(self.pos, size)
        # print("read =>", r0)
        for r1 in self.ranges:
            if not r1.overlap(r0):
                yield None, r1
                continue

            # [r0 [ ... r1] never happens as the first r1 always starts with 0
            assert r1.start <= r0.start

            # [r1 [ r0 ] ] r1 cached; prevent unnecessary area split
            if r1.includes(r0) and r1.cached:
                data = os.pread(self.cachefp.fileno(), r0.length, r0.start)
                # print(r1, 'includes', r0, 'len(r0)?=', len(data))
                yield data, r1
                continue

            yield None, _Range(r1.start, r0.start - r1.start, r1.cached)

            # [ r1 [ r0 ] ]
            if r0.right < r1.right:
                yield self._get_range(_Range(r0.start, r0.length, r1.cached))
                # print('[ r1 [ r0 ] ]', r, len(data))
                yield None, _Range(r0.right, r1.right - r0.right, r1.cached)
                continue

            # [ r1 [ ] r0 ]
            yield self._get_range(_Range(r0.start, r1.right - r0.start,
                                         r1.cached))
            # print('[ r1 [ ] r0 ]', r, len(data))

            r0 = _Range(r1.right, r0.right - r1.right)

    def _get_range(self, r) -> (bytearray, _Range):
        # print('get range:', r)
        if r.cached:
            return os.pread(self.cachefp.fileno(), r.length, r.start), r

        assert not self._frozen
        self.fileobj.seek(r.start, io.SEEK_SET)
        data = self.fileobj.read(r.length)
        written = os.pwrite(self.cachefp.fileno(), data, r.start)
        if written < 0:
            raise RuntimeError("bad file descriptor")
        # print(written, "/", r.length, "bytes written at", r.start)
        return data, _Range(r.start, r.length, True)


class CachedWrapper(_CachedWrapperBase):
    '''A page-based transparent local cache for remote files

    This wrapper makes a transparent read-only local cache as sparse
    file.  The local cache behaves as read-mirror of remote file -
    when a known range is requested, it'll be local read. If it's not
    locally cached, it fetches the range and stores as local file.

    .. note:: It's not thread-safe yet.

    Example usage follows:

    .. code-block::

        from pfio.v2 import from_url
        from pfio.cache import SparseFileCache

        with from_url("s3://bucket/path-prefix/") as s3:
          large_file = "path/to/large/file"
          stat = s3.stat(large_file)
          with SparseFileCache(s3.open(large_file), stat.size,
                               close_on_close=True) as fp:
            fp.seek(1024)
            # Read file from 1024 to 1024+65536 and cache it
            data = fp.read(65536)


    It is internally used behind ZIP fs:

    .. code-block::

        from pfio.v2 import from_url

        with from_url("s3://bucket/your.zip", local_cache=True) as fs:
          with fs.open("file-in-zip.jpg", 'rb') as fp:
            data = rp.read()

    '''

    def __init__(self, fileobj, size, cachedir=None, close_on_close=False,
                 pagesize=16*1024*1024, multithread_safe=False,
                 cache_size_limit=None):
        super().__init__(fileobj, size, cachedir, close_on_close,
                         cache_size_limit=cache_size_limit,
                         multithread_safe=multithread_safe)
        assert pagesize > 0
        self.pagesize = pagesize
        self.size = size
        self._init_ranges()

    def _init_ranges(self):
        self.cache_size = 0
        pagecount = self.size // self.pagesize
        self.ranges = [_Range(i * self.pagesize, self.pagesize, cached=False)
                       for i in range(pagecount)]

        remain = self.size % self.pagesize
        if remain > 0:
            r = _Range(pagecount*self.pagesize, remain, cached=False)
            self.ranges.append(r)

    def read(self, size=-1) -> bytes:
        with self.lock.wrlock():
            # TODO: it this giant lock becomes the bottleneck, split
            # this lock into per-page locks
            return self._read(size)

    def _is_full(self) -> bool:
        if self.cache_size_limit is None:
            return False
        return self.cache_size >= self.cache_size_limit

    def _read(self, size) -> bytes:
        if self._closed:
            raise RuntimeError("closed")

        if size < 0 or (self.size - self.pos < size):
            size = self.size - self.pos

        start = self.pos // self.pagesize
        end = (self.pos + size) // self.pagesize
        if (self.pos + size) % self.pagesize != 0:
            end += 1

        for i in range(start, end):
            # print('range=', i, "total=", len(self.ranges))
            r = self.ranges[i]

            if not r.cached:
                assert not self._frozen
                self.fileobj.seek(r.start, io.SEEK_SET)
                # print("fetching", r.start, r.length, os.getpid())
                data = self.fileobj.read(r.length)
                n = os.pwrite(self.cachefp.fileno(), data, r.start)
                if n < 0:
                    raise RuntimeError("bad file descriptor")

                self.ranges[i] = _Range(r.start, r.length, cached=True)

        buf = os.pread(self.cachefp.fileno(), size, self.pos)

        self.pos += len(buf)
        self.pos %= self.size
        if self.pos != self.fileobj.tell():
            self.fileobj.seek(self.pos, io.SEEK_SET)

        # cache_size_limit is a soft limit
        if self._is_full():
            os.truncate(self.cachefp.fileno(), 0)
            self._init_ranges()

        return buf


@contextmanager
def _shflock(fd):
    fcntl.flock(fd, fcntl.LOCK_SH)
    try:
        yield
    finally:
        fcntl.flock(fd, fcntl.LOCK_UN)


@contextmanager
def _exflock(fd):
    fcntl.flock(fd, fcntl.LOCK_EX)
    try:
        yield
    finally:
        fcntl.flock(fd, fcntl.LOCK_UN)


class MPCachedWrapper(CachedWrapper):
    '''Multiprocessing version of cached wrapper

        It is internally used behind ZIP fs for multiprocessing:

    .. code-block::

        import os
        from pfio.v2 import from_url

        with from_url("s3://bucket/your.zip",
                      local_cache=True, reset_on_fork=True) as fs:

          pid = os.fork()
          if pid:
            os.wait()
          else:
            with fs.open("file-in-zip.jpg", 'rb') as fp:
              data = rp.read()
    '''

    def __init__(self, fileobj, size, cachedir=None, close_on_close=False,
                 cache_size_limit=None, pagesize=16*1024*1024,
                 local_cachefile=None, local_indexfile=None,
                 multithread_safe=False):
        super().__init__(fileobj, size, cachedir, close_on_close,
                         cache_size_limit=cache_size_limit,
                         multithread_safe=multithread_safe)
        assert pagesize > 0
        self.pagesize = pagesize
        self.size = size

        # Both none or both string file
        assert bool(local_indexfile) == bool(local_cachefile)

        if local_indexfile is None:
            self.indexfp = tempfile.NamedTemporaryFile(delete=False,
                                                       dir=self.cachedir)
            self.cachefp = tempfile.NamedTemporaryFile(delete=False,
                                                       dir=self.cachedir)
            self.local_indexfile = self.indexfp.name
            self.local_cachefile = self.cachefp.name
            self._open_fd(self.indexfp.name, self.cachefp.name)
            self._init_indexfile()

        else:
            self.indexfp = open(local_indexfile, 'rb')
            self.cachefp = open(local_cachefile, 'rb')

            self._open_fd(local_indexfile, local_cachefile)

        self.pid = os.getpid()

    def close(self):
        if self._closed:
            return

        os.close(self.cachefd)
        os.close(self.indexfd)
        self.indexfp.close()
        self.cachefp.close()
        self._closed = True

    def _open_fd(self, indexfile, cachefile):
        self.local_indexfile = indexfile
        self.local_cachefile = cachefile
        self.indexfd = os.open(self.local_indexfile, os.O_RDWR)
        self.cachefd = os.open(self.local_cachefile, os.O_RDWR)

    def _init_indexfile(self):
        pagecount = self.size // self.pagesize

        for i in range(pagecount):
            r = _Range(i * self.pagesize, self.pagesize, cached=False)
            self._set_index(i, r)
            r = self._get_index(i)

        remain = self.size % self.pagesize
        if remain > 0:
            offset = pagecount * self.pagesize
            r = _Range(offset, remain, cached=False)
            self._set_index(pagecount, r)

    def _get_index(self, i):
        assert i >= 0
        width = _Range.size()
        buf = os.pread(self.indexfd, width, width * i)
        return _Range.unpack(buf)

    def _set_index(self, i, r):
        assert i >= 0
        assert r is not None
        width = _Range.size()
        written = os.pwrite(self.indexfd, r.pack(), width * i)
        assert width == written

    def read(self, size=-1) -> bytes:
        if self._closed:
            raise RuntimeError("closed")

        if size < 0 or (self.size - self.pos < size):
            size = self.size - self.pos

        start = self.pos // self.pagesize
        end = (self.pos + size) // self.pagesize
        if (self.pos + size) % self.pagesize != 0:
            end += 1

        for i in range(start, end):
            with _shflock(self.indexfd):
                r = self._get_index(i)

            if not r.cached:
                assert not self._frozen

                with _exflock(self.indexfd):
                    r = self._get_index(i)
                    if r.cached:
                        continue

                    self.fileobj.seek(r.start, io.SEEK_SET)
                    # print("fetching", r.start, r.length, os.getpid())
                    data = self.fileobj.read(r.length)
                    n = os.pwrite(self.cachefd, data, r.start)
                    if n < 0:
                        raise RuntimeError("bad file descriptor")
                    self._set_index(i, _Range(r.start, r.length, cached=True))

        buf = os.pread(self.cachefd, size, self.pos)

        self.pos += len(buf)
        self.pos %= self.size
        if self.pos != self.fileobj.tell():
            self.fileobj.seek(self.pos, io.SEEK_SET)

        with _shflock(self.indexfd):
            is_full = self._is_full()

        # If the cache file is more than the limit, just flush them all.
        if is_full:
            with _exflock(self.indexfd):
                is_full = self._is_full()
                if is_full:
                    os.truncate(self.cachefd, 0)
                    self._init_indexfile()

        return buf

    def _is_full(self):
        if self.cache_size_limit is None:
            return False
        stat = os.stat(self.local_cachefile)
        return stat.st_blocks * 512 >= self.cache_size_limit



import errno
import numbers
import os
import pickle
import subprocess
import tempfile
import threading
import warnings
from struct import calcsize, pack, unpack

from pfio import cache

# Deprecated, but leaving for backward compatibility just in case any
# system directly using this value
_DEFAULT_CACHE_PATH = os.path.join(
    os.getenv('HOME'), ".pfio", "cache")

_FORCE_LOCAL = True


def _default_cache_path():
    basedir = os.getenv('XDG_CACHE_HOME')

    if not basedir:
        basedir = os.path.join(os.getenv('HOME'), ".cache")

    cachedir = os.path.join(basedir, "pfio")
    return cachedir


class LockContext:
    def __init__(self, locked_lock):
        self.locked_lock = locked_lock

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.locked_lock.unlock()


class RWLock:
    '''Reader-writer lock

    TODO(kuenishi): Add unit tests

    '''

    def __init__(self):
        self.cv = threading.Condition()
        self.writer = None
        self.reader = set()

    def rdlock(self):
        with self.cv:
            self.cv.wait_for(lambda: self.writer is None)
            self.reader.add(threading.get_ident())
            return LockContext(self)

    def wrlock(self):
        with self.cv:
            thread_id = threading.get_ident()
            self.cv.wait_for(lambda: self.writer is None and
                             self.writer != thread_id and
                             len(self.reader) == 0)
            self.writer = thread_id
            return LockContext(self)

    def unlock(self):
        with self.cv:
            thread_id = threading.get_ident()
            if self.writer == thread_id:
                self.writer = None
            else:
                self.reader.remove(thread_id)
            self.cv.notify_all()


class DummyLock:
    '''Dummy class for multithread-unsafe fast cache class
    '''

    def __init__(self):
        pass

    def rdlock(self):
        return LockContext(self)

    def wrlock(self):
        return LockContext(self)

    def unlock(self):
        pass


def _check_local(path: str):
    '''Verifies the path is not NFS

    Raises an error if the ``stat(1)`` answers it as NFS.

    '''
    global _FORCE_LOCAL
    if not _FORCE_LOCAL:
        return

    # From https://unix.stackexchange.com/a/402236
    cp = subprocess.run(['stat', '-f', '-c', '%T', path], check=True,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT)
    if 'nfs' in cp.stdout.decode():
        raise ValueError("NFS can't be used for cache: {}".format(path))


class FileCache(cache.Cache):
    '''Cache system with local filesystem

    Stores cache data in a local temporary file created in
    ``$XDG_CACHE_HOME/pfio`` by default. If it is unset,
    ``$HOME/.cache/pfio`` will be the cache destination. Cache data is
    automatically deleted after the object is collected. When this
    object is not correctly closed, (e.g., the process killed by
    SIGTERM), the cache remains after the death of process.

    .. note:: This feature requires ``stat(1)`` command from GNU
       coreutils.

    Arguments:
        length (int): Length of the cache array.

        multithread_safe (bool): Defines multithread safety. If this
            is ``True``, reader-writer locking system based on
            ``threading.Lock`` is introduced behind the cache
            management. Major use case is with Chainer's
            ``MultithreadIterator``.

        do_pickle (bool):
            Do automatic pickle and unpickle inside the cache.

        dir (str): The path to the directory to place cache data in
            case home directory is not backed by fast storage device.
            Must not be an NFS.

        cache_size_limit (None or int): Limitation of the cache size in bytes.
            If the total amount of cached data reaches the limit, the cache
            will become frozen and no longer acccept further addition.
            Data already stored in the cache can be accessed normally.
            None (default) and 0 is unlimited.

        verbose (bool):
            Print detailed logs of the cache.

    '''

    def __init__(self, length, multithread_safe=False, do_pickle=False,
                 dir=None, cache_size_limit=None, verbose=False):
        self._multithread_safe = multithread_safe
        self.length = length
        self.do_pickle = do_pickle
        if self.length <= 0 or (2 ** 64) <= self.length:
            raise ValueError("length has to be between 0 and 2^64")

        if not (cache_size_limit is None or
                (isinstance(cache_size_limit, numbers.Number) and
                 0 <= cache_size_limit)):
            msg = "cache_size_limit has to be either None, zero " \
                  "(both indicate unlimited) or larger than 0. " \
                  "{} is specified.".format(cache_size_limit)
            raise ValueError(msg)
        self.cache_size_limit = cache_size_limit

        if self.multithread_safe:
            self.lock = RWLock()
        else:
            self.lock = DummyLock()

        if dir is None:
            self.dir = _default_cache_path()
        else:
            self.dir = dir
        os.makedirs(self.dir, exist_ok=True)
        _check_local(self.dir)

        self.closed = False
        self.cachefp = tempfile.NamedTemporaryFile(delete=True, dir=self.dir)

        # allocate space to store 2n uint64 index buffer filled by -1.
        # the cache data will be appended after the indices.
        buf = pack('Qq', 0, -1)
        self.buflen = calcsize('Qq')
        assert self.buflen == 16
        for i in range(self.length):
            offset = self.buflen * i
            r = os.pwrite(self.cachefp.fileno(), buf, offset)
            assert r == self.buflen
        self.pos = self.buflen * self.length

        self.verbose = verbose
        if self.verbose:
            print('created cache file:', self.cachefp.name)

        self._frozen = False

    def __len__(self):
        return self.length

    @property
    def frozen(self):
        return self._frozen

    @property
    def multiprocess_safe(self):
        # If it's preseved/preloaded, then the file contents are fixed.
        return self._frozen

    @property
    def multithread_safe(self):
        return self._multithread_safe

    def get(self, i):
        if self.closed:
            return
        data = self._get(i)
        if self.do_pickle and data:
            data = pickle.loads(data)
        return data

    def _get(self, i):
        if i < 0 or self.length <= i:
            raise IndexError("index {} out of range ([0, {}])"
                             .format(i, self.length - 1))

        offset = self.buflen * i
        with self.lock.rdlock():
            buf = os.pread(self.cachefp.fileno(), self.buflen, offset)
            (o, l) = unpack('Qq', buf)
            if l < 0 or o < 0:
                return None

            data = os.pread(self.cachefp.fileno(), l, o)
            assert len(data) == l
            return data

    def put(self, i, data):
        if self._frozen or self.closed:
            return False

        try:
            if self.do_pickle:
                data = pickle.dumps(data)
            return self._put(i, data)

        except OSError as ose:
            # Disk full (ENOSPC) possibly by cache; just warn and keep running
            if ose.errno == errno.ENOSPC:
                warnings.warn(ose.strerror, RuntimeWarning)
                return False
            else:
                raise ose

    def _put(self, i, data):
        if self.closed:
            return
        if i < 0 or self.length <= i:
            raise IndexError("index {} out of range ([0, {}])"
                             .format(i, self.length - 1))

        if self.cache_size_limit:
            if self.cache_size_limit < (self.pos + len(data)):
                self._frozen = True
                return False

        offset = self.buflen * i
        with self.lock.wrlock():
            buf = os.pread(self.cachefp.fileno(), self.buflen, offset)
            (o, l) = unpack('Qq', buf)
            if l >= 0 and o >= 0:
                # Already data exists
                return False

            pos = self.pos

            '''Notes on possibility of partial write

            write(3) says partial writes ret<nbyte may happen in
            case nbytes>PIPE_BUF. In Linux 5.0 PIPE_BUF is
            4096 so partial writes do not happen when writing
            index, but they may happen when writing data. We
            hope it is rare, it seems to happen mostly in case
            of multiple writer processes, disk full and
            ``EINTR``.

            CPython does care this case by retrying
            ``pwrite(2)`` as long as it returns ``-1`` . But
            returns when the return value is positive. We'd
            better care that case.

            '''
            buf = pack('Qq', pos, len(data))
            r = os.pwrite(self.cachefp.fileno(), buf, offset)
            assert r == self.buflen

            current_pos = pos
            while current_pos - pos < len(data):
                r = os.pwrite(self.cachefp.fileno(),
                              data[current_pos-pos:], current_pos)
                assert r > 0
                current_pos += r
            assert current_pos - pos == len(data)

            self.pos += len(data)
            return True

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self):
        with self.lock.wrlock():
            if not self.closed:
                self.closed = True
                self.cachefp.close()
                self.cachefp = None

    def preload(self, name):
        '''Load the cache saved by ``preserve()``

        ``cache_path`` is the path to the persistent file. To use cache
        in ``multiprocessing`` environment, call this method at every
        forked process, except the process that called ``preserve()``.
        After the preload, no data can be added to the cache.

        When it succeeds, it returns ``True``.
        If there is no cache file with the specified name in
        the cache directory, it will do nothing but return ``False``.

        Returns:
            bool: Returns True if succeed.

        .. note:: This feature is experimental.

        '''
        if self._frozen:
            if self.verbose:
                print("Failed to preload the cache from {}: "
                      "The cache is already frozen."
                      .format(name))
            return False

        cachefile = os.path.join(self.dir, name)

        if not os.path.exists(cachefile):
            if self.verbose:
                print('Failed to ploread the cache from {}: '
                      'The specified cache not found in {}'
                      .format(name, self.dir))
            return False

        with self.lock.wrlock():
            self.cachefp.close()
            self.cachefp = open(cachefile, 'rb')
            self._frozen = True
        return True

    def preserve(self, name, overwrite=False):
        '''Preserve the cache as a persistent file on the disk

        Saves the current cache into ``cache_path``.
        Once the cache is preserved, the cache file will not be removed
        at cache close. To read data from the preserved file, use
        ``preload()`` method. After preservation, no data can be added
        to the cache.

        When it succeeds, it returns ``True``.
        If there is a cache file with the same name already exists in the
        cache directory, it will do nothing but return ``False``.

        The preserved cache can also be preloaded by
        :class:`~MultiprocessFileCache`.

        Arguments:
            name (str): Prefix of the preserved file names.
                ``(name).cachei`` and ``(name).cached`` are created.
                The files are created in the same directory as the cache
                (``dir`` option to ``__init__``).

            overwrite (bool): Overwrite if already exists

        Returns:
            bool: Returns True if succeed.

        .. note:: This feature is experimental.

        '''

        cachefile = os.path.join(self.dir, name)

        if overwrite:
            if os.path.exists(cachefile):
                os.unlink(cachefile)
        elif os.path.exists(cachefile):
            if self.verbose:
                print('Specified cache named "{}" already exists in {}'
                      .format(name, self.dir))
            return False

        with self.lock.wrlock():
            # Hard link and save them
            os.link(self.cachefp.name, cachefile)
            self.cachefp.close()

            self.cachefp = open(cachefile, 'rb')
            self._frozen = True
        return True



import abc
from abc import abstractmethod
from typing import Callable, Optional


class Cache(abc.ABC):
    '''Abstract class to define Cache class interface

    This can be instance of ``collections.abc.Sequence`` but so far
    this will be just a single interface difinition.  Note that this
    is experimental feature.

    '''

    def __init__(self):
        pass

    @abstractmethod
    def __len__(self) -> int:
        "Returns the length of the cache data"
        raise NotImplementedError()

    @property
    @abstractmethod
    def multiprocess_safe(self) -> bool:
        "Returns multiprocess safety."
        raise NotImplementedError()

    @property
    @abstractmethod
    def multithread_safe(self) -> bool:
        "Returns multithread safety."
        raise NotImplementedError()

    @abstractmethod
    def put(self, i: int, data: bytes) -> bool:
        "Puts the bytes to the cache. No overwrite nor deletion supported."
        raise NotImplementedError()

    @abstractmethod
    def get(self, i: int) -> Optional[bytes]:
        "Tries to get the data from cache."
        raise NotImplementedError()

    def get_and_cache(self, i, backend_get: Callable[[int], bytes]) -> bytes:
        '''Get data from cache, otherwise from backend with caching

        First try to get data from cache. If not found, it gets data
        from backend callable with the result stored in cache.

        '''
        data = self.get(i)
        if data is None:
            data = backend_get(i)
            self.put(i, data)
        return data


from pfio.cache.file_cache import FileCache  # NOQA
from pfio.cache.mmap_file_cache import ReadOnlyFileCache  # NOQA
from pfio.cache.multiprocess_file_cache import MultiprocessFileCache  # NOQA
from pfio.cache.naive import NaiveCache  # NOQA
from pfio.cache.sparse_file import CachedWrapper as SparseFileCache  # NOQA
from pfio.cache.sparse_file import \
    MPCachedWrapper as MultiprocessSparseFileCache  # NOQA



import abc
import contextlib
import copy
import os
import stat
from abc import abstractmethod
from io import IOBase
from types import TracebackType
from typing import Any, Callable, Iterator, Optional, Type
from urllib.parse import urlparse

from deprecation import deprecated

from pfio.version import __version__  # NOQA


class FileStat(abc.ABC):
    """Detailed file or directory information abstraction

    :meth:`pfio.v2.FS.stat` returns an object that implements of
    ``FileStat``.  In addition to the common attributes that the
    ``FileStat`` abstract provides, each ``FileStat`` subclass
    implements some additional attributes depending on what
    information the corresponding filesystem or container can handle.
    The common attributes have the same behavior despite filesystem or
    container type difference.

    Attributes:
        filename (str):
            Filename in the filesystem or container.
        last_modifled (float):
            UNIX timestamp of mtime. Note that some
            filesystems or containers do not have sub-second precision.
        mode (int):
            Permission with file type flag (regular file or directory).
            You can make a human-readable interpretation by
            `stat.filemode <https://docs.python.org/3/library/stat.html#stat.filemode>`_.
        size (int):
            Size in bytes. Note that directories may have different
            sizes depending on the filesystem or container type.

    """     # NOQA
    filename = None
    last_modified = None
    mode = None
    size = None

    def isdir(self):
        """Returns whether the target is a directory, based on the permission flag

        Note that some systems do not support directory tree semantics.

        Returns:
            `True` if directory, `False` otherwise.
        """
        return bool(self.mode & 0o40000)

    def __str__(self):
        if isinstance(self.mode, int):
            mode = stat.filemode(self.mode)
        else:
            mode = self.mode
        return '<{} filename="{}" mode="{}">'.format(
            type(self).__name__, self.filename, mode)

    def __repr__(self):
        return str(self.__str__())


class ForkedError(RuntimeError):
    '''An error class when PFIO found the process forked.

    If an FS object is not "lazy", any object usage detects process
    fork and raises this ``ForkedError`` as soon as possible at the
    child process. The parent process may or may not run well,
    depending on the ``FS`` implementation.

    '''
    pass


class FS(abc.ABC):
    '''FS access abstraction

    '''

    _cwd = ''

    def __init__(self, reset_on_fork=False):
        self.pid = os.getpid()
        self.reset_on_fork = reset_on_fork

    @property
    def cwd(self):
        return self._cwd

    @cwd.setter
    def cwd(self, value):
        self._cwd = value

    @abstractmethod
    def open(self, file_path: str, mode: str = 'rb',
             buffering: int = -1, encoding: Optional[str] = None,
             errors: Optional[str] = None,
             newline: Optional[str] = None,
             closefd: bool = True,
             opener: Optional[Callable[
                 [str, int], Any]] = None) -> Type["IOBase"]:
        raise NotImplementedError()

    def open_zip(self, file_path: str, mode='r', **kwargs) -> Type["Zip"]:  # NOQA
        from .zip import Zip
        return Zip(self, file_path, mode, **kwargs)

    def subfs(self, rel_path: str) -> Type["FS"]:
        '''Virtually changes the working directory

        By default it performs shallow copy. If any resource that as
        different lifecycles than the copy source (e.g. HDFS
        connection and zipfile.ZipFile object), they also will be
        copied by overriding this method.

        '''
        if rel_path.startswith("/"):
            raise RuntimeError("Absolute path is not supported")
        elif '..' in rel_path.split(os.path.sep):
            raise RuntimeError("Only subtree is supported")

        sub = copy.copy(self)

        sub._cwd = os.path.join(self.cwd, rel_path)
        sub._reset()
        return sub

    def _checkfork(self):
        if not self.is_forked:
            return

        # Forked!
        if self.reset_on_fork:
            self._reset()
            self.pid = os.getpid()
        else:
            raise ForkedError()

    @abstractmethod
    def _reset(self):
        raise NotImplementedError()

    @property
    def is_forked(self):
        assert hasattr(self, 'pid')
        return self.pid != os.getpid()

    def close(self) -> None:
        pass

    @abstractmethod
    def list(self, path_or_prefix: Optional[str] = None,
             recursive=False) -> Iterator[str]:
        """Lists all the files and directories under
           the given ``path_or_prefix``

        Args:
            path_or_prefix (str): The path to list against.
                When we get the default value, ``list`` shows the content under
                the working directory as the default value.
                However, if a ``path_or_prefix`` is given,
                then it shows only the files and directories
                under the ``path_or_prefix``.

            recursive (bool): When this is ``True``, list files and directories
                recursively.

        Returns:
            An Iterator that iterates though the files and directories.

        """
        raise NotImplementedError()

    @abstractmethod
    def stat(self, path: str) -> FileStat:
        """Show details of a file

        It returns an object of subclass of :class:`pfio.io.FileStat`
        in accordance with filesystem or container type.

        Args:
            path (str): The path to file

        Returns:
            :class:`pfio.io.FileStat` object.
        """
        raise NotImplementedError()

    def __enter__(self) -> 'FS':
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        self.close()

    @abstractmethod
    def isdir(self, file_path: str) -> bool:
        """Returns ``True`` if the path is an existing directory

        Args:
            path (str): the path to the target directory

        Returns:
            ``True`` when the path points to a directory,
            ``False`` when it is not

        """
        raise NotImplementedError()

    @abstractmethod
    def mkdir(self, file_path: str, mode: int = 0o777,
              *args, dir_fd: Optional[int] = None) -> None:
        """Makes a directory with mode

        Args:
            path (str): the path to the directory to make

            mode (int): the mode of the new directory

        """
        raise NotImplementedError()

    @abstractmethod
    def makedirs(self, file_path: str, mode: int = 0o777,
                 exist_ok: bool = False) -> None:
        """Makes directories recursively with mode

        Also creates all the missing parents of the given path.

        Args:
            path (str): the path to the directory to make.

            mode (int): the mode of the directory

            exist_ok (bool): In default case, a ``FileExitsError`` will be
                raised when the target directory exists.

        """
        raise NotImplementedError()

    @abstractmethod
    def exists(self, path: str) -> bool:
        """Returns the existence of the path

        When the ``file_path`` points to a symlink, the return value
        depends on the actual file instead of the link itself.

        """
        raise NotImplementedError()

    @abstractmethod
    def rename(self, src: str, dst: str) -> None:
        """Renames the file from ``src`` to ``dst``

        On systems and situation where rename functionality is
        proviced, it renames the file or the directory.


        Args:
            src (str): the current name of the file or directory.

            dst (str): the name to rename to.

        """
        raise NotImplementedError()

    @abstractmethod
    def remove(self, file_path: str, recursive: bool = False) -> None:
        """Removes a file or directory

           Args:
               path (str): the target path to remove. The ``path`` can be a
               regular file or a directory.

               recursive (bool): When the given path is a directory,
                   all the files and directories under it will be removed.
                   When the path is a file, this option is ignored.

        """
        raise NotImplementedError()


@contextlib.contextmanager
def open_url(url: str, mode: str = 'r', **kwargs) -> 'IOBase':
    '''Opens a file regardless of the backend FS type

    ``url`` must be compliant with URL standard in
    https://url.spec.whatwg.org/ .  As this function implements
    context manager, the FileObject can be written as::

       with open_url("s3://bucket.example.com/path/your-file.txt", 'r') as f:
           f.read()

    .. note:: Some FS resouces won't be closed when using this
        functionality. See ``from_url`` for keyword arguments.

    Returns:
        a FileObject that must be closed.

    '''
    dirname, filename = os.path.split(url)
    with from_url(dirname, **kwargs) as fs:
        with fs.open(filename, mode) as fp:
            yield fp


def from_url(url: str, **kwargs) -> 'FS':
    '''Factory pattern implementation, creates FS from URI

    If ``force_type`` is set with archive type, not scheme,
    it ignores the suffix and tries the specified archive
    format by opening the blob file.

    If ``force_type`` is set with scheme type, the FS will
    built from it accordingly. The URL path is supposed to
    be a directory for file systems or a path prefix for S3.

    .. warning:: When opening an ``hdfs://...`` URL, be sure about
        forking context. See: :class:`Hdfs` for discussion.

    Arguments:
        url (str): A URL string compliant with RFC 1738.

        force_type (str): Force type of FS to be returned.
            One of "zip", "hdfs", "s3", or "file", returned
            respectively. Default is ``"file"``.

        create (bool): Create the specified path doesn't exist.

        reset_on_fork (bool): Reset stateful objects (e.g. connection
            to the remote system) after fork.

    .. note:: Some FS resouces won't be closed when using this
        functionality.

    .. note:: Pickling the FS object may or may not work correctly
        depending on the implementation.

    '''
    parsed = urlparse(url)
    force_type = kwargs.pop('force_type', None)

    if parsed.scheme:
        scheme = parsed.scheme
    else:
        scheme = 'file'  # Default is local

    # When ``force_type`` is defined, it must be equal with given one.
    if force_type is not None and force_type != "zip":
        if force_type != scheme:
            raise ValueError("URL scheme mismatch with forced type")

    def _zip_check_create_not_supported():
        if kwargs.get('create', False):
            msg = '"create" option is not supported for Zip FS.'
            raise ValueError(msg)

    # force_type \ suffix | .zip    | other
    # --------------------+---------+------
    #                 zip | ok      | try zip
    #             (other) | try dir | try dir
    #                None | try zip | try dir
    if force_type == 'zip':
        _zip_check_create_not_supported()
        dirname, filename = os.path.split(parsed.path)
        fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)
        fs = fs.open_zip(filename, **kwargs)

    elif force_type is None:
        if parsed.path.endswith('.zip'):
            _zip_check_create_not_supported()
            dirname, filename = os.path.split(parsed.path)
            fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)
            fs = fs.open_zip(filename, **kwargs)
        else:
            dirname = parsed.path
            fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)

    else:
        dirname = parsed.path
        fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)

    return fs


def _from_scheme(scheme, dirname, kwargs, bucket=None):
    if scheme == 'file':
        from .local import Local
        fs = Local(dirname, **kwargs)
    elif scheme == 'hdfs':
        from .hdfs import Hdfs
        fs = Hdfs(dirname, **kwargs)
    elif scheme == 's3':
        from .s3 import S3
        fs = S3(bucket=bucket, prefix=dirname, **kwargs)
    else:
        raise RuntimeError("Scheme {} is not supported", scheme)

    return fs


@deprecated(deprecated_in='2.2.0', removed_in='2.3.0',
            current_version=__version__)
def lazify(init_func, lazy_init=True, recreate_on_fork=True) -> "FS":
    '''Make FS init lazy and recreate on fork

    '''
    return _LazyFS(init_func, lazy_init, recreate_on_fork)


class _LazyFS:
    def __init__(self, func, lazy_init, recreate_on_fork):
        if not lazy_init:
            self.mixin = func()
        else:
            self.mixin = None
        self.func = func

    def __getattr__(self, attr):
        if self.mixin is None or self.mixin.is_forked:
            self.mixin = self.func()
        return getattr(self.mixin, attr)

    def __enter__(self) -> 'FS':
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> bool:
        self.close()



# -*- coding: utf-8

import os

from setuptools import find_packages, setup

package_data = []

proj_dir = os.path.dirname(__file__)
templates = os.path.join(proj_dir, 'resources', 'templates')

for root, dirs, names in os.walk(templates):
    for fname in names:
        abspath = os.path.join(root, fname)
        relpath = os.path.relpath(proj_dir, abspath)
        package_data.append(relpath)
        print(relpath)

here = os.path.abspath(os.path.dirname(__file__))
# Get __version__ variable
exec(open(os.path.join(here, 'pfio', 'version.py')).read())

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pfio',
    version=__version__,
    description='PFN IO library',
    author='Tianqi Xu, Kota Uenishi',
    author_email='tianqi@preferred.jp, kota@preferred.jp',
    url='http://github.com/pfnet/pfio',
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',

        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',

        'Topic :: System :: Filesystems',
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    package_data={'pfio': package_data},
    extras_require={
        'test': ['pytest', 'flake8', 'autopep8', 'parameterized', 'isort', 'moto', 'numpy'],
        # When updating doc deps, docs/requirements.txt should be updated too
        'doc': ['sphinx', 'sphinx_rtd_theme'],
        'bench': ['numpy>=1.19.5', 'torch>=1.9.0', 'Pillow<=8.2.0'],
    },
    python_requires=">=3.6",
    # When updating install requires, docs/requirements.txt should be updated too
    install_requires=['pyarrow>=6.0.0', 'boto3', 'deprecation'],
    include_package_data=True,
    zip_safe=False,

    keywords='filesystem hdfs chainer development',
)



import fnmatch
import os
from pathlib import PurePath

from .fs import FS
from .local import Local


class Path:
    sep = '/'

    def __init__(self, *args, fs=None, root=None):
        '''

        Empty argument indicates current directory in fs, fs.cwd

        '''
        self._fs = fs if fs else Local()
        assert isinstance(self._fs, FS)

        self._root = root

        parts = []
        for argv in args:
            if argv.startswith(self.sep):
                parts = []
            parts.append(argv)
        if len(parts) == 0 and root is None:
            parts = [self._fs.cwd]

        self._parts = parts

        if self._parts and self._parts[0] == self.sep and self._root is None:
            self._parts = self._parts[1:]
            self._root = self.sep

    @property
    def name(self):
        filename = self._root if not self._parts else self._parts[-1]
        return os.path.normpath(filename)

    @property
    def suffix(self):
        return os.path.splitext(self.resolve())[1]

    def with_suffix(self, ext):
        assert ext.startswith('.')
        parts = self._parts[:]
        if parts:
            base = os.path.splitext(parts[-1])[0]
            parts[-1] = '{}{}'.format(base, ext)
            p = Path(*parts, fs=self._fs, root=self._root)
            return p

        base = os.path.splitext(self._root)[0]
        root = '{}{}'.format(base, ext)
        p = Path(*parts, fs=self._fs, root=root)
        return p

    @property
    def parent(self):
        path = str(self)
        parent = os.path.split(path)[0]
        return Path(parent, fs=self._fs, root=self._root)

    def __rtruediv__(self, a):
        if isinstance(a, str):
            lhs = Path(a, fs=self._fs, root=self._root)
        else:
            lhs = a
        return lhs / self

    def __truediv__(self, a):
        parts = self._parts[:]
        if isinstance(a, Path):
            parts += a._parts
        elif isinstance(a, PurePath):
            raise RuntimeError("mixed")
        else:
            parts.append(str(a))

        p = Path(*parts, fs=self._fs, root=self._root)
        return p

    @classmethod
    def cwd(cls):
        raise NotImplementedError("cwd() is unsupported on this system")

    @classmethod
    def home(cls):
        raise NotImplementedError("home() is unsupported on this system")

    def exists(self):
        return self._fs.exists(self.resolve())

    def is_absolute(self):
        return self._root is not None

    def __repr__(self):
        return "{}[{}:{}]".format(self.__class__.__name__,
                                  self._fs.__class__.__name__,
                                  str(self))

    def __str__(self):

        if self._root:
            return os.path.join(self._root, *self._parts)

        return os.path.join(*self._parts)

    def __fspath__(self):
        return str(self)

    def __lt__(self, other):
        return str(self) < str(other)

    def resolve(self, strict=True):
        base = self._root if self._root else self._fs.cwd
        if '..' in self._parts or '.' in self._parts:
            raise RuntimeError("TODO")
        parts = self._parts[:]
        return Path(*parts, fs=self._fs, root=base)

    def samefile(self, other):
        return str(self.resolve()) == str(other.resolve())

    def touch(self):
        with self._fs.open(self.resolve(), 'wb') as fp:
            fp.write(b'')

    def open(self, mode='r', **kwargs):
        # TODO: handle or warn on ignoring these keyword arguments
        # buffering=-1, encoding=None, errors=None, newline=None):
        return self._fs.open(self.resolve(), mode)

    def mkdir(self, mode=0o777, parents=False, exist_ok=False):
        if parents:
            return self._fs.makedirs(self.resolve(), mode, exist_ok)

        try:
            self._fs.mkdir(self.resolve(), mode)
        except FileExistsError as e:
            if not exist_ok:
                raise e
            else:
                pass

    def is_dir(self):
        return self._fs.isdir(self.resolve())

    def is_file(self):
        return not self._fs.isdir(self.resolve())

    def glob(self, pattern):
        '''glob

        TODO: this implemention is not smart enough in case of large
        subdicrectories as they won't be skipped regarding the input
        pattern.

        Should return ``Path`` class as well.

        '''
        base = self.resolve()
        pattern1 = os.path.join(base, pattern)

        for p in self._fs.list(base, recursive=True):
            if fnmatch.fnmatch(os.path.join(base, p), pattern1):
                yield Path(p, fs=self._fs, root=self._root)

    def stat(self):
        return self._fs.stat(self.resolve())

    def unlink(self, missing_ok=False):
        return self._fs.remove(self.resolve())

    def rename(self, target):
        raise NotImplementedError("rename() is unsupported on this system")

    def read_text(self):
        with self.open() as fp:
            return fp.read()

    def read_bytes(self):
        with self.open(mode='rb') as fp:
            return fp.read()

    def write_text(self, data):
        with self.open(mode='w') as fp:
            return fp.write(data)

    def write_bytes(self, data):
        view = memoryview(data)
        with self.open(mode='wb') as fp:
            return fp.write(view)
